"""Shared constants and backend-agnostic helpers for metric extraction."""

from __future__ import annotations

import math
from typing import Any, Protocol

TASK_CLASSIFICATION = "classification"
TASK_REGRESSION = "regression"
RIDGE_LAMBDA = 1e-6
NUM_BOOTSTRAP = 200
EPS = 1e-12


class Shaped(Protocol):
    """Structural protocol for array/tensor-like inputs with `.shape`."""

    @property
    def shape(self) -> Any:
        """Return shape payload."""


def finite_or_none(value: float) -> float | None:
    """Return finite values as float; map non-finite values to None."""

    return float(value) if math.isfinite(value) else None


def resolve_task_from_metadata(metadata: dict[str, Any]) -> str | None:
    """Resolve task directly from metadata config payload when present."""

    config = metadata.get("config")
    if isinstance(config, dict):
        dataset_cfg = config.get("dataset")
        if isinstance(dataset_cfg, dict):
            raw_task = dataset_cfg.get("task")
            if raw_task in {TASK_CLASSIFICATION, TASK_REGRESSION}:
                return str(raw_task)
    return None


def validate_metric_shapes(
    *,
    feature_types: list[str],
    x_train: Shaped,
    x_test: Shaped,
    y_train: Shaped,
    y_test: Shaped,
) -> None:
    """Validate metric extraction array/tensor shapes and feature metadata size."""

    if int(x_train.shape[1]) != int(x_test.shape[1]):
        raise ValueError(
            f"X_train/X_test feature mismatch: {int(x_train.shape[1])} != {int(x_test.shape[1])}"
        )
    if int(x_train.shape[0]) != int(y_train.shape[0]):
        raise ValueError(
            f"X_train/y_train row mismatch: {int(x_train.shape[0])} != {int(y_train.shape[0])}"
        )
    if int(x_test.shape[0]) != int(y_test.shape[0]):
        raise ValueError(
            f"X_test/y_test row mismatch: {int(x_test.shape[0])} != {int(y_test.shape[0])}"
        )
    if int(x_train.shape[0]) + int(x_test.shape[0]) <= 0:
        raise ValueError("Dataset must contain at least one row across train and test splits.")
    if len(feature_types) != int(x_train.shape[1]):
        raise ValueError(
            "feature_types length must match feature count: "
            f"{len(feature_types)} != {int(x_train.shape[1])}"
        )



"""Tests for functions/multi.py."""

import pytest
import torch
from conftest import make_generator as _make_generator
from conftest import make_keyed_rng as _make_keyed_rng

from dagzoo.core.fixed_layout.batched import FixedLayoutBatchRng, apply_function_plan_batch
from dagzoo.core.fixed_layout.plan_types import (
    ConcatNodeSource,
    GaussianMatrixPlan,
    LinearFunctionPlan,
    StackedNodeSource,
)
from dagzoo.core.layout_types import AggregationKind
from dagzoo.functions.multi import apply_multi_function
from dagzoo.functions.random_functions import apply_random_function
from dagzoo.math import sanitize_and_standardize


def test_single_input() -> None:
    g = _make_generator(0)
    x = torch.randn(64, 4, generator=g)
    y = apply_multi_function([x], g, out_dim=5)
    assert y.shape == (64, 5)


def test_single_input_matches_apply_random_function() -> None:
    x = torch.randn(64, 4, generator=_make_generator(3))
    actual_generator = _make_generator(4)
    reference_generator = _make_generator(4)

    actual = apply_multi_function([x.clone()], actual_generator, out_dim=5)
    expected = apply_random_function(x.clone(), reference_generator, out_dim=5)

    torch.testing.assert_close(actual, expected)
    torch.testing.assert_close(actual_generator.get_state(), reference_generator.get_state())


def test_multiple_inputs() -> None:
    g = _make_generator(1)
    a = torch.randn(64, 3, generator=g)
    b = torch.randn(64, 2, generator=g)
    y = apply_multi_function([a, b], g, out_dim=4)
    assert y.shape == (64, 4)


@pytest.mark.parametrize("aggregation_kind", ["sum", "product", "max", "logsumexp"])
def test_multiple_inputs_support_explicit_aggregation_kind(
    aggregation_kind: AggregationKind,
) -> None:
    g = _make_generator(2)
    a = torch.randn(64, 3, generator=g)
    b = torch.randn(64, 2, generator=g)
    y = apply_multi_function([a, b], g, out_dim=4, aggregation_kind=aggregation_kind)
    assert y.shape == (64, 4)
    assert torch.all(torch.isfinite(y))


@pytest.mark.parametrize("aggregation_kind", ["sum", "product", "max", "logsumexp"])
def test_multiple_inputs_are_deterministic_for_explicit_aggregation_kind(
    aggregation_kind: AggregationKind,
) -> None:
    a = torch.randn(64, 3, generator=_make_generator(11))
    b = torch.randn(64, 2, generator=_make_generator(12))
    y1 = apply_multi_function(
        [a.clone(), b.clone()],
        _make_generator(13),
        out_dim=4,
        aggregation_kind=aggregation_kind,
    )
    y2 = apply_multi_function(
        [a.clone(), b.clone()],
        _make_generator(13),
        out_dim=4,
        aggregation_kind=aggregation_kind,
    )
    torch.testing.assert_close(y1, y2)


def test_multiple_inputs_concat_sanitizes_non_finite_values(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    source = ConcatNodeSource(function=LinearFunctionPlan(matrix=GaussianMatrixPlan()))
    monkeypatch.setattr(
        "dagzoo.functions.multi.sample_multi_source_plan", lambda *_args, **_kwargs: source
    )
    inputs = [
        torch.tensor(
            [[0.0, float("nan")], [1.0, 2.0], [float("inf"), -1.0]],
            dtype=torch.float32,
        ),
        torch.tensor([[1.0], [2.0], [-float("inf")]], dtype=torch.float32),
    ]

    actual = apply_multi_function(inputs, _make_generator(18), out_dim=3)

    assert actual.shape == (3, 3)
    assert torch.all(torch.isfinite(actual))


def test_multiple_inputs_stacked_sanitizes_non_finite_values(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    source = StackedNodeSource(
        aggregation_kind="sum",
        parent_functions=(
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
        ),
    )
    monkeypatch.setattr(
        "dagzoo.functions.multi.sample_multi_source_plan", lambda *_args, **_kwargs: source
    )
    inputs = [
        torch.tensor(
            [[0.0, float("nan")], [1.0, 2.0], [float("inf"), -1.0]],
            dtype=torch.float32,
        ),
        torch.tensor([[1.0], [2.0], [-float("inf")]], dtype=torch.float32),
    ]

    actual = apply_multi_function(inputs, _make_generator(19), out_dim=3)

    assert actual.shape == (3, 3)
    assert torch.all(torch.isfinite(actual))


def test_multi_function_matches_explicit_stacked_plan(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    inputs = [
        torch.randn(64, 3, generator=_make_generator(21)),
        torch.randn(64, 2, generator=_make_generator(22)),
        torch.randn(64, 4, generator=_make_generator(23)),
    ]
    source = StackedNodeSource(
        aggregation_kind="logsumexp",
        parent_functions=(
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
        ),
    )
    monkeypatch.setattr(
        "dagzoo.functions.multi.sample_multi_source_plan", lambda *_args, **_kwargs: source
    )

    actual_generator = _make_generator(24)
    reference_generator = _make_generator(24)

    actual = apply_multi_function(
        [inp.clone() for inp in inputs],
        actual_generator,
        out_dim=5,
    )
    root = _make_keyed_rng(reference_generator, "apply_multi_function")
    rng = FixedLayoutBatchRng.from_keyed_rng(root.keyed("execution"), batch_size=1, device="cpu")
    transformed = [
        apply_function_plan_batch(
            sanitize_and_standardize(inp).unsqueeze(0),
            rng.keyed("parent", plan_index),
            source.parent_functions[plan_index],
            out_dim=5,
            noise_sigma_multiplier=1.0,
            noise_spec=None,
            standardize_input=False,
        ).squeeze(0)
        for plan_index, inp in enumerate(inputs)
    ]
    expected = torch.logsumexp(torch.stack(transformed, dim=1), dim=1)

    torch.testing.assert_close(actual, expected)
    torch.testing.assert_close(actual_generator.get_state(), reference_generator.get_state())


@pytest.mark.parametrize("aggregation_kind", ["sum", "product", "max"])
def test_multi_function_matches_explicit_stacked_plan_for_associative_reducers(
    aggregation_kind: AggregationKind,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    inputs = [
        torch.randn(64, 3, generator=_make_generator(31)),
        torch.randn(64, 2, generator=_make_generator(32)),
        torch.randn(64, 4, generator=_make_generator(33)),
    ]
    source = StackedNodeSource(
        aggregation_kind=aggregation_kind,
        parent_functions=(
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
            LinearFunctionPlan(matrix=GaussianMatrixPlan()),
        ),
    )
    monkeypatch.setattr(
        "dagzoo.functions.multi.sample_multi_source_plan", lambda *_args, **_kwargs: source
    )

    actual_generator = _make_generator(34)
    reference_generator = _make_generator(34)
    actual = apply_multi_function(
        [inp.clone() for inp in inputs],
        actual_generator,
        out_dim=5,
    )

    root = _make_keyed_rng(reference_generator, "apply_multi_function")
    rng = FixedLayoutBatchRng.from_keyed_rng(root.keyed("execution"), batch_size=1, device="cpu")
    transformed = [
        apply_function_plan_batch(
            sanitize_and_standardize(inp).unsqueeze(0),
            rng.keyed("parent", plan_index),
            source.parent_functions[plan_index],
            out_dim=5,
            noise_sigma_multiplier=1.0,
            noise_spec=None,
            standardize_input=False,
        ).squeeze(0)
        for plan_index, inp in enumerate(inputs)
    ]
    expected = torch.stack(transformed, dim=1)
    if aggregation_kind == "sum":
        expected_out = torch.sum(expected, dim=1)
    elif aggregation_kind == "product":
        expected_out = torch.prod(expected, dim=1)
    else:
        expected_out = torch.max(expected, dim=1).values

    torch.testing.assert_close(actual, expected_out)
    torch.testing.assert_close(actual_generator.get_state(), reference_generator.get_state())


def test_empty_raises() -> None:
    g = _make_generator(0)
    with pytest.raises(ValueError, match="non-empty"):
        apply_multi_function([], g, out_dim=3)


def test_deterministic() -> None:
    x = torch.randn(32, 4, generator=_make_generator(99))
    y1 = apply_multi_function([x.clone()], _make_generator(0), out_dim=3)
    y2 = apply_multi_function([x.clone()], _make_generator(0), out_dim=3)
    torch.testing.assert_close(y1, y2)


def test_finite_outputs() -> None:
    g = _make_generator(7)
    x = torch.randn(64, 4, generator=g)
    y = apply_multi_function([x], g, out_dim=3)
    assert torch.all(torch.isfinite(y))



"""Dataset generation entrypoints."""

from .dataset import (
    generate_batch,
    generate_batch_iter,
    generate_one,
)

__all__ = [
    "generate_batch",
    "generate_batch_iter",
    "generate_one",
]



#!/usr/bin/env python3
"""Convenience wrapper for `dagzoo diversity-audit`."""

from __future__ import annotations

import sys
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from dagzoo.cli import main as cli_main  # noqa: E402


def main(argv: list[str] | None = None) -> int:
    args = sys.argv[1:] if argv is None else argv
    return cli_main(["diversity-audit", *args])


if __name__ == "__main__":
    raise SystemExit(main())



"""Shared availability guards for temporarily unsupported filtering workflows."""

from __future__ import annotations

from typing import NoReturn

FILTERING_UNSUPPORTED_MESSAGE = (
    "Deferred filtering is temporarily disabled; generated outputs are the only "
    "supported corpus artifact for now."
)


def raise_filtering_unsupported() -> NoReturn:
    """Raise the shared unsupported-feature error for filtering workflows."""

    raise NotImplementedError(FILTERING_UNSUPPORTED_MESSAGE)



from __future__ import annotations

import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[2]
SRC_ROOT = REPO_ROOT / "src" / "dagzoo"
DOCS_DEP_MAP_PATH = REPO_ROOT / "docs" / "development" / "module-dependency-map.md"


@dataclass(frozen=True)
class CommandSpec:
    label: str
    argv: tuple[str, ...]


class DevToolError(RuntimeError):
    """Raised for developer tooling failures with user-facing messages."""


def repo_relative(path: Path) -> str:
    try:
        return path.relative_to(REPO_ROOT).as_posix()
    except ValueError:
        return path.as_posix()


def python_tool(tool_name: str) -> str:
    candidate = REPO_ROOT / ".venv" / "bin" / tool_name
    if candidate.exists():
        return str(candidate)
    return tool_name


def venv_python() -> Path:
    return REPO_ROOT / ".venv" / "bin" / "python"


def run_command(command: CommandSpec) -> None:
    result = subprocess.run(command.argv, cwd=REPO_ROOT, check=False)
    if result.returncode != 0:
        raise DevToolError(f"{command.label} failed with exit code {result.returncode}.")


def format_command(argv: tuple[str, ...]) -> str:
    return " ".join(argv)


def tool_exists(tool_name: str) -> bool:
    return shutil.which(tool_name) is not None



"""Deferred filter replay helpers."""

from __future__ import annotations

import math
from collections.abc import Mapping
from typing import Any

from dagzoo.config import FilterConfig
from dagzoo.filter_thresholds import validate_filter_threshold
from dagzoo.rng import SEED32_MAX, SEED32_MIN


def _coerce_seed(raw_seed: object, *, dataset_index: int) -> int:
    """Resolve a valid seed32 for filter replay."""

    if isinstance(raw_seed, bool):
        raw_seed = None

    if isinstance(raw_seed, float):
        if math.isfinite(raw_seed) and float(raw_seed).is_integer():
            raw_seed = int(raw_seed)
        else:
            raw_seed = None

    if isinstance(raw_seed, int):
        if SEED32_MIN <= raw_seed <= SEED32_MAX:
            return int(raw_seed)

    return int(dataset_index % (SEED32_MAX + 1))


def _resolve_filter_seed(metadata_payload: Mapping[str, Any], *, dataset_index: int) -> int:
    """Resolve filter replay seed from persisted metadata with child-seed preference."""

    return _coerce_seed(
        metadata_payload.get("dataset_seed", metadata_payload.get("seed")),
        dataset_index=dataset_index,
    )


def _resolve_task_and_filter_config(
    *,
    metadata_payload: Mapping[str, Any],
    threshold_override: float | None,
    n_jobs_override: int | None,
) -> tuple[str, FilterConfig]:
    """Resolve task + filter config for one dataset record."""

    task: str | None = None
    embedded_filter: Mapping[str, Any] | None = None

    config_payload = metadata_payload.get("config")
    if isinstance(config_payload, Mapping):
        dataset_payload = config_payload.get("dataset")
        if isinstance(dataset_payload, Mapping):
            dataset_task = dataset_payload.get("task")
            if isinstance(dataset_task, str) and dataset_task.strip():
                task = dataset_task.strip().lower()

        filter_payload = config_payload.get("filter")
        if isinstance(filter_payload, Mapping):
            embedded_filter = filter_payload

    if task not in {"classification", "regression"}:
        raise ValueError(
            "Deferred filter requires embedded metadata.config.dataset.task in shard metadata."
        )

    if embedded_filter is not None:
        filter_cfg = FilterConfig(**dict(embedded_filter))
    else:
        raise ValueError(
            "Deferred filter requires embedded metadata.config.filter in shard metadata."
        )

    filter_cfg.enabled = True
    if threshold_override is not None:
        filter_cfg.threshold = validate_filter_threshold(
            threshold_override,
            field_name="threshold_override",
        )
    if n_jobs_override is not None:
        filter_cfg.n_jobs = int(n_jobs_override)
        filter_cfg.__post_init__()

    return task, filter_cfg


def _build_filter_metadata(
    *,
    existing_filter: object,
    accepted: bool,
    filter_details: Mapping[str, Any],
) -> dict[str, Any]:
    """Build normalized filter metadata payload for deferred status."""

    payload: dict[str, Any] = dict(existing_filter) if isinstance(existing_filter, Mapping) else {}
    payload["mode"] = "deferred"
    payload["status"] = "accepted" if accepted else "rejected"
    payload["enabled"] = True
    payload["accepted"] = bool(accepted)
    payload.update(dict(filter_details))
    return payload



import json

import pytest

from dagzoo.config import GeneratorConfig
from dagzoo.diagnostics.effective_diversity import (
    resolve_filter_calibration_thresholds,
    run_filter_calibration,
    validate_filter_calibration_threshold,
    write_filter_calibration_artifacts,
)


def _audit_entry(
    *,
    label: str,
    threshold_effective_mean: float,
    filter_accepted_datasets_per_minute: float,
    diversity_acceptance_rate: float,
) -> dict[str, object]:
    return {
        "label": label,
        "config_path": "configs/preset_filter_benchmark_smoke.yaml",
        "suite": "smoke",
        "num_datasets": 10,
        "warmup_datasets": 0,
        "requested_device": "cpu",
        "resolved_device": "cpu",
        "resolved_config": {"filter": {"enabled": True}},
        "datasets_per_minute": 100.0,
        "filter_datasets_per_minute": 50.0,
        "filter_accepted_datasets_per_minute": float(filter_accepted_datasets_per_minute),
        "filter_accepted_datasets_measured": 4,
        "filter_rejected_datasets_measured": 6,
        "filter_acceptance_rate_dataset_level": float(diversity_acceptance_rate),
        "filter_rejection_rate_dataset_level": float(1.0 - diversity_acceptance_rate),
        "mechanism_family_summary": {
            "metadata_coverage_rate": 1.0,
            "bundles_with_metadata": 10,
            "sampled_family_counts": {"linear": 10},
            "dataset_presence_rate_by_family": {"linear": 1.0},
            "sampled_variant_counts": {},
            "dataset_presence_rate_by_variant": {},
            "mean_total_function_plans": 4.0,
        },
        "filter_summary": {
            "accepted_true_fraction": float(diversity_acceptance_rate),
            "wins_ratio_mean": 0.91,
            "threshold_effective_mean": float(threshold_effective_mean),
            "threshold_delta_mean": 0.0,
            "n_valid_oob_mean": 96.0,
            "reason_counts": {"insufficient_oob_predictions": 1},
        },
    }


def test_resolve_filter_calibration_thresholds_uses_default_offsets_and_baseline() -> None:
    assert resolve_filter_calibration_thresholds(baseline_threshold=0.95, thresholds=None) == [
        0.8,
        0.85,
        0.9,
        0.95,
        1.0,
    ]


def test_resolve_filter_calibration_thresholds_clamps_defaults_into_zero_to_one() -> None:
    assert resolve_filter_calibration_thresholds(baseline_threshold=0.05, thresholds=None) == [
        0.0,
        0.05,
        0.1,
    ]


def test_resolve_filter_calibration_thresholds_dedupes_explicit_override() -> None:
    assert resolve_filter_calibration_thresholds(
        baseline_threshold=0.95,
        thresholds=[1.0, 0.8, 0.95, 1.0],
    ) == [0.8, 0.95, 1.0]


@pytest.mark.parametrize("value", [float("nan"), float("inf"), float("-inf"), -0.1, 1.1])
def test_validate_filter_calibration_threshold_rejects_invalid_values(value: float) -> None:
    with pytest.raises(ValueError, match=r"must be a finite value in \[0.0, 1.0\]"):
        validate_filter_calibration_threshold(value, field_name="filter.threshold")


@pytest.mark.parametrize("value", [float("nan"), float("inf"), float("-inf"), -0.1, 1.1])
def test_run_filter_calibration_rejects_invalid_baseline_config_threshold(
    value: float,
) -> None:
    cfg = GeneratorConfig.from_yaml("configs/preset_filter_benchmark_smoke.yaml")
    cfg.filter.threshold = value

    with pytest.raises(
        ValueError, match=r"filter.threshold must be a finite value in \[0.0, 1.0\]"
    ):
        run_filter_calibration(
            config=cfg,
            config_path="configs/preset_filter_benchmark_smoke.yaml",
            thresholds=None,
            suite="smoke",
            num_datasets=10,
            warmup=0,
            device="cpu",
            warn_threshold_pct=2.5,
            fail_threshold_pct=5.0,
        )


def test_run_filter_calibration_ranks_best_overall_and_best_passing(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cfg = GeneratorConfig.from_yaml("configs/preset_filter_benchmark_smoke.yaml")
    captured: dict[str, object] = {}

    def _stub_run_effective_diversity_audit(**kwargs):
        captured["variant_labels"] = kwargs["variant_labels"]
        return {
            "generated_at": "2026-03-10T08:00:00+00:00",
            "baseline": _audit_entry(
                label="baseline",
                threshold_effective_mean=0.95,
                filter_accepted_datasets_per_minute=40.0,
                diversity_acceptance_rate=0.4,
            ),
            "variants": [
                _audit_entry(
                    label="thr_0.8",
                    threshold_effective_mean=0.8,
                    filter_accepted_datasets_per_minute=60.0,
                    diversity_acceptance_rate=0.6,
                ),
                _audit_entry(
                    label="thr_1.0",
                    threshold_effective_mean=1.0,
                    filter_accepted_datasets_per_minute=80.0,
                    diversity_acceptance_rate=0.3,
                ),
            ],
            "comparisons": [
                {
                    "variant_label": "thr_0.8",
                    "diversity_status": "pass",
                    "diversity_composite_shift_pct": 1.5,
                    "diversity_metric_shift_pct": {"linearity_proxy": 1.5},
                    "datasets_per_minute_delta_pct": 5.0,
                    "filter_accepted_datasets_per_minute_delta_pct": 50.0,
                },
                {
                    "variant_label": "thr_1.0",
                    "diversity_status": "fail",
                    "diversity_composite_shift_pct": 8.0,
                    "diversity_metric_shift_pct": {"linearity_proxy": 8.0},
                    "datasets_per_minute_delta_pct": 0.0,
                    "filter_accepted_datasets_per_minute_delta_pct": 100.0,
                },
            ],
            "summary": {
                "overall_status": "fail",
                "probe_num_datasets": 10,
                "probe_warmup_datasets": 0,
            },
        }

    monkeypatch.setattr(
        "dagzoo.diagnostics.effective_diversity.calibration.run_effective_diversity_audit",
        _stub_run_effective_diversity_audit,
    )

    report = run_filter_calibration(
        config=cfg,
        config_path="configs/preset_filter_benchmark_smoke.yaml",
        thresholds=[0.8, 1.0],
        suite="smoke",
        num_datasets=10,
        warmup=0,
        device="cpu",
        warn_threshold_pct=2.5,
        fail_threshold_pct=5.0,
    )

    assert captured["variant_labels"] == ["thr_0.8", "thr_1.0"]
    assert report["schema_name"] == "dagzoo_filter_calibration_report"
    assert report["summary"]["baseline_threshold_requested"] == pytest.approx(0.95)
    assert report["summary"]["best_overall_threshold_requested"] == pytest.approx(1.0)
    assert report["summary"]["best_overall_diversity_status"] == "fail"
    assert report["summary"]["best_passing_threshold_requested"] == pytest.approx(0.8)
    assert report["candidates"][0]["diversity_status"] == "pass"
    assert report["baseline"]["diversity_status"] == "reference"
    assert report["candidates"][1]["threshold_effective_mean"] == pytest.approx(0.95)


def test_run_filter_calibration_keeps_fine_grained_threshold_candidates_distinct(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cfg = GeneratorConfig.from_yaml("configs/preset_filter_benchmark_smoke.yaml")
    captured: dict[str, object] = {}

    def _stub_run_effective_diversity_audit(**kwargs):
        captured["variant_labels"] = kwargs["variant_labels"]
        return {
            "generated_at": "2026-03-10T08:00:00+00:00",
            "baseline": _audit_entry(
                label="baseline",
                threshold_effective_mean=0.95,
                filter_accepted_datasets_per_minute=40.0,
                diversity_acceptance_rate=0.4,
            ),
            "variants": [
                _audit_entry(
                    label="thr_0.801",
                    threshold_effective_mean=0.801,
                    filter_accepted_datasets_per_minute=61.0,
                    diversity_acceptance_rate=0.61,
                ),
                _audit_entry(
                    label="thr_0.804",
                    threshold_effective_mean=0.804,
                    filter_accepted_datasets_per_minute=54.0,
                    diversity_acceptance_rate=0.54,
                ),
            ],
            "comparisons": [
                {
                    "variant_label": "thr_0.801",
                    "diversity_status": "pass",
                    "diversity_composite_shift_pct": 1.0,
                    "diversity_metric_shift_pct": {"linearity_proxy": 1.0},
                    "datasets_per_minute_delta_pct": 2.0,
                    "filter_accepted_datasets_per_minute_delta_pct": 52.5,
                },
                {
                    "variant_label": "thr_0.804",
                    "diversity_status": "warn",
                    "diversity_composite_shift_pct": 3.5,
                    "diversity_metric_shift_pct": {"linearity_proxy": 3.5},
                    "datasets_per_minute_delta_pct": 1.0,
                    "filter_accepted_datasets_per_minute_delta_pct": 35.0,
                },
            ],
            "summary": {
                "overall_status": "warn",
                "probe_num_datasets": 10,
                "probe_warmup_datasets": 0,
            },
        }

    monkeypatch.setattr(
        "dagzoo.diagnostics.effective_diversity.calibration.run_effective_diversity_audit",
        _stub_run_effective_diversity_audit,
    )

    report = run_filter_calibration(
        config=cfg,
        config_path="configs/preset_filter_benchmark_smoke.yaml",
        thresholds=[0.801, 0.804],
        suite="smoke",
        num_datasets=10,
        warmup=0,
        device="cpu",
        warn_threshold_pct=2.5,
        fail_threshold_pct=5.0,
    )

    assert captured["variant_labels"] == ["thr_0.801", "thr_0.804"]
    assert [candidate["label"] for candidate in report["candidates"][:2]] == [
        "thr_0.801",
        "thr_0.804",
    ]
    assert report["candidates"][0]["filter_accepted_datasets_per_minute"] == pytest.approx(61.0)
    assert report["candidates"][0]["diversity_status"] == "pass"
    assert report["candidates"][1]["filter_accepted_datasets_per_minute"] == pytest.approx(54.0)
    assert report["candidates"][1]["diversity_status"] == "warn"
    assert report["summary"]["best_overall_threshold_requested"] == pytest.approx(0.801)
    assert report["summary"]["best_passing_threshold_requested"] == pytest.approx(0.801)


def test_write_filter_calibration_artifacts(tmp_path) -> None:
    report = {
        "schema_name": "dagzoo_filter_calibration_report",
        "schema_version": 1,
        "baseline": {
            "label": "baseline",
            "threshold_requested": 0.95,
            "filter_accepted_datasets_per_minute": 45.0,
            "filter_acceptance_rate_dataset_level": 0.5,
            "mechanism_family_summary": {
                "metadata_coverage_rate": 1.0,
                "bundles_with_metadata": 10,
                "sampled_family_counts": {"gp": 10, "linear": 10},
                "dataset_presence_rate_by_family": {"gp": 1.0, "linear": 1.0},
                "sampled_variant_counts": {"gp.multiscale": 10},
                "dataset_presence_rate_by_variant": {"gp.multiscale": 1.0},
                "mean_total_function_plans": 6.0,
            },
        },
        "candidates": [
            {
                "label": "baseline",
                "threshold_requested": 0.95,
                "diversity_status": "reference",
                "filter_accepted_datasets_per_minute": 45.0,
                "filter_acceptance_rate_dataset_level": 0.5,
                "diversity_composite_shift_pct": None,
                "mechanism_family_summary": {
                    "metadata_coverage_rate": 1.0,
                    "bundles_with_metadata": 10,
                    "sampled_family_counts": {"gp": 10, "linear": 10},
                    "dataset_presence_rate_by_family": {"gp": 1.0, "linear": 1.0},
                    "sampled_variant_counts": {"gp.multiscale": 10},
                    "dataset_presence_rate_by_variant": {"gp.multiscale": 1.0},
                    "mean_total_function_plans": 6.0,
                },
            }
        ],
        "comparisons": [],
        "summary": {
            "overall_status": "reference",
            "baseline_threshold_requested": 0.95,
            "best_overall_threshold_requested": 0.95,
            "best_overall_diversity_status": "reference",
            "best_passing_threshold_requested": None,
            "num_candidates": 1,
            "probe_num_datasets": 10,
            "probe_warmup_datasets": 0,
        },
    }

    artifact_paths = write_filter_calibration_artifacts(report, out_dir=tmp_path)
    payload = json.loads(artifact_paths["summary_json"].read_text(encoding="utf-8"))
    markdown = artifact_paths["summary_md"].read_text(encoding="utf-8")

    assert payload["schema_name"] == "dagzoo_filter_calibration_report"
    assert "Best overall threshold" in markdown
    assert "canonical persisted artifacts for filter calibration" in markdown
    assert "## Mechanism Families" in markdown
    assert "gp.multiscale" in markdown



"""Random activations."""

from __future__ import annotations

import torch

from dagzoo.functions._rng_helpers import rand_scalar, randint_scalar
from dagzoo.math import (
    log_uniform as _log_uniform,
)
from dagzoo.math import (
    standardize as _standardize,
)

_FIXED_ACTIVATION_NAMES: tuple[str, ...] = (
    "tanh",
    "leaky_relu",
    "elu",
    "identity",
    "silu",
    "swiglu",
    "relu",
    "relu_sq",
    "softplus",
    "sign",
    "gauss",
    "exp",
    "sin",
    "square",
    "abs",
    "softmax",
    "logsigmoid",
    "logabs",
    "sigmoid",
    "round",
    "mod1",
    "selu",
    "relu6",
    "hardtanh",
    "indicator_01",
    "onehot_argmax",
    "argsort",
    "rank",
)
_GUMBEL_SOFTMAX_EPS = 1e-6


def fixed_activation_names() -> tuple[str, ...]:
    """Return the canonical fixed-activation sampling order."""

    return _FIXED_ACTIVATION_NAMES


def _gumbel_softmax_activation(
    x: torch.Tensor,
    *,
    temperature: float,
    uniform_noise: torch.Tensor,
    dim: int = -1,
) -> torch.Tensor:
    """Apply one Gumbel-softmax relaxation using provided uniform noise."""

    safe_temperature = max(float(temperature), 1e-3)
    u = uniform_noise.to(device=x.device, dtype=x.dtype)
    u = torch.clamp(u, min=_GUMBEL_SOFTMAX_EPS, max=1.0 - _GUMBEL_SOFTMAX_EPS)
    gumbel = -torch.log(-torch.log(u))
    logits = (x + gumbel) / safe_temperature
    return torch.softmax(logits, dim=dim)


def _fixed_activation(x: torch.Tensor, name: str) -> torch.Tensor:
    """Apply one fixed activation variant by name in torch."""
    if name == "tanh":
        return torch.tanh(x)
    if name == "leaky_relu":
        return torch.nn.functional.leaky_relu(x, negative_slope=0.01)
    if name == "elu":
        return torch.nn.functional.elu(x)
    if name == "identity":
        return x
    if name == "silu":
        return torch.nn.functional.silu(x)
    if name == "swiglu":
        return x * torch.nn.functional.silu(x)
    if name == "relu":
        return torch.relu(x)
    if name == "relu_sq":
        return torch.square(torch.relu(x))
    if name == "softplus":
        return torch.nn.functional.softplus(x)
    if name == "sign":
        return torch.sign(x)
    if name == "gauss":
        return torch.exp(-(x**2))
    if name == "exp":
        return torch.exp(torch.clamp(x, max=20.0))
    if name == "sin":
        return torch.sin(x)
    if name == "square":
        return x * x
    if name == "abs":
        return torch.abs(x)
    if name == "softmax":
        return torch.softmax(x, dim=1)
    if name == "logsigmoid":
        return torch.nn.functional.logsigmoid(x)
    if name == "logabs":
        return torch.log(torch.clamp(torch.abs(x), min=1e-6))
    if name == "sigmoid":
        return torch.sigmoid(x)
    if name == "round":
        return torch.round(x)
    if name == "mod1":
        return torch.remainder(x, 1.0)
    if name == "selu":
        return torch.nn.functional.selu(x)
    if name == "relu6":
        return torch.nn.functional.relu6(x)
    if name == "hardtanh":
        return torch.nn.functional.hardtanh(x)
    if name == "indicator_01":
        return ((x >= 0) & (x <= 1)).to(x.dtype)
    if name == "onehot_argmax":
        indices = torch.argmax(x, dim=1)
        return torch.nn.functional.one_hot(indices, num_classes=x.shape[1]).to(x.dtype)
    if name == "argsort":
        return torch.argsort(x, dim=1).to(x.dtype)
    if name == "rank":
        return (torch.argsort(torch.argsort(x, dim=1), dim=1) + 1).to(x.dtype)
    return x


def _param_activation(x: torch.Tensor, generator: torch.Generator) -> torch.Tensor:
    """Apply one randomly parameterized activation family in torch."""
    choices = ["relu_pow", "signed_pow", "inv_pow", "poly", "gumbel_softmax"]
    idx = randint_scalar(0, len(choices), generator)
    choice = choices[int(idx)]

    if choice == "relu_pow":
        q = _log_uniform(generator, 0.1, 10.0, str(x.device))
        return torch.pow(torch.clamp(x, min=0.0), q)
    if choice == "signed_pow":
        q = _log_uniform(generator, 0.1, 10.0, str(x.device))
        return torch.sign(x) * torch.pow(torch.abs(x), q)
    if choice == "inv_pow":
        q = _log_uniform(generator, 0.1, 10.0, str(x.device))
        return torch.pow(torch.abs(x) + 1e-3, -q)
    if choice == "gumbel_softmax":
        temperature = _log_uniform(generator, 0.25, 4.0, str(x.device))
        uniform_noise = torch.rand(
            x.shape,
            generator=generator,
            device=x.device,
            dtype=x.dtype,
        )
        return _gumbel_softmax_activation(
            x,
            temperature=temperature,
            uniform_noise=uniform_noise,
            dim=1,
        )

    m = randint_scalar(2, 6, generator)
    return torch.pow(x, float(m))


def apply_random_activation(
    x: torch.Tensor,
    generator: torch.Generator,
    *,
    with_standardize: bool = True,
    parametric_prob: float = 1.0 / 3.0,
) -> torch.Tensor:
    """Apply one random activation from supported families in torch."""
    y = x.to(torch.float32)
    if y.dim() == 1:
        y = y.unsqueeze(1)

    device = str(y.device)

    if with_standardize:
        y = _standardize(y)
        a = _log_uniform(generator, 1.0, 10.0, device)
        row_idx = randint_scalar(0, y.shape[0], generator)
        b = y[int(row_idx) : int(row_idx) + 1]
        y = a * (y - b)

    if rand_scalar(generator) < parametric_prob:
        y = _param_activation(y, generator)
    else:
        fixed = fixed_activation_names()
        idx = randint_scalar(0, len(fixed), generator)
        y = _fixed_activation(y, fixed[int(idx)])

    y = torch.nan_to_num(y, nan=0.0, posinf=1e6, neginf=-1e6)
    y = torch.clamp(y, -1e6, 1e6)

    if with_standardize:
        y = _standardize(y)
    return y.to(torch.float32)



from __future__ import annotations

import ast
import json
from dataclasses import dataclass
from pathlib import Path

from .common import DOCS_DEP_MAP_PATH, REPO_ROOT, SRC_ROOT, repo_relative

HOTSPOT_MODULES = (
    "dagzoo.core.execution_semantics",
    "dagzoo.core.fixed_layout.batched",
    "dagzoo.core.fixed_layout.runtime",
    "dagzoo.core.layout",
    "dagzoo.core.node_pipeline",
    "dagzoo.core.dataset",
    "dagzoo.core.config_resolution",
    "dagzoo.core.generation_runtime",
    "dagzoo.config",
    "dagzoo.io.lineage_schema",
)


@dataclass(frozen=True)
class ModuleSummary:
    imports: tuple[str, ...]
    direct_importers: tuple[str, ...]
    transitive_importers: tuple[str, ...]
    impacted_packages: tuple[str, ...]


@dataclass(frozen=True)
class ImportGraph:
    imports: dict[str, tuple[str, ...]]
    reverse_imports: dict[str, tuple[str, ...]]
    module_paths: dict[str, str]

    @property
    def modules(self) -> tuple[str, ...]:
        return tuple(sorted(self.imports))

    def package_graph(self) -> dict[str, set[str]]:
        graph: dict[str, set[str]] = {}
        for module, imports in self.imports.items():
            package = module_to_package(module)
            graph.setdefault(package, set())
            for dependency in imports:
                dependency_package = module_to_package(dependency)
                if dependency_package != package:
                    graph[package].add(dependency_package)
                    graph.setdefault(dependency_package, set())
        return graph

    def transitive_importers(self, module: str) -> tuple[str, ...]:
        visited: set[str] = set()
        stack = list(self.reverse_imports.get(module, ()))
        while stack:
            current = stack.pop()
            if current in visited:
                continue
            visited.add(current)
            stack.extend(self.reverse_imports.get(current, ()))
        return tuple(sorted(visited))

    def module_summary(self, module: str) -> ModuleSummary:
        direct_importers = self.reverse_imports.get(module, ())
        transitive_importers = self.transitive_importers(module)
        impacted_packages = sorted(
            {
                module_to_package(importer)
                for importer in transitive_importers
                if module_to_package(importer)
                not in {module_to_package(module), "dagzoo", "dagzoo.__main__"}
            }
        )
        return ModuleSummary(
            imports=self.imports.get(module, ()),
            direct_importers=direct_importers,
            transitive_importers=transitive_importers,
            impacted_packages=tuple(impacted_packages),
        )

    def module_to_path(self, module: str) -> str | None:
        return self.module_paths.get(module)


def module_to_package(module: str) -> str:
    parts = module.split(".")
    if len(parts) <= 2:
        return module
    return ".".join(parts[:2])


def module_name_for_path(path: Path) -> str:
    relative = path.relative_to(SRC_ROOT)
    module = "dagzoo." + ".".join(relative.parts)
    module = module[:-3] if module.endswith(".py") else module
    if module.endswith(".__init__"):
        module = module[:-9]
    return module


def path_to_module(path_str: str) -> str | None:
    path = REPO_ROOT / path_str
    if not path.is_relative_to(SRC_ROOT) or path.suffix != ".py":
        return None
    return module_name_for_path(path)


def build_import_graph() -> ImportGraph:
    module_paths: dict[str, str] = {}
    known_modules: set[str] = set()
    for path in sorted(SRC_ROOT.rglob("*.py")):
        if "__pycache__" in path.parts:
            continue
        module = module_name_for_path(path)
        module_paths[module] = repo_relative(path)
        known_modules.add(module)

    imports: dict[str, tuple[str, ...]] = {}
    reverse_imports: dict[str, set[str]] = {module: set() for module in known_modules}
    for module, path_str in module_paths.items():
        path = REPO_ROOT / path_str
        module_imports = tuple(sorted(_collect_imports(path, module, known_modules)))
        imports[module] = module_imports
        for dependency in module_imports:
            reverse_imports.setdefault(dependency, set()).add(module)

    return ImportGraph(
        imports=imports,
        reverse_imports={key: tuple(sorted(value)) for key, value in reverse_imports.items()},
        module_paths=module_paths,
    )


def _collect_imports(path: Path, module: str, known_modules: set[str]) -> set[str]:
    tree = ast.parse(path.read_text())
    current_package = module if path.name == "__init__.py" else module.rsplit(".", 1)[0]
    discovered: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                normalized = _normalize_import(alias.name, known_modules)
                if normalized is not None and normalized != module:
                    discovered.add(normalized)
        elif isinstance(node, ast.ImportFrom):
            for dependency in _normalize_import_from(node, current_package, known_modules):
                if dependency != module:
                    discovered.add(dependency)
    return discovered


def _normalize_import(candidate: str, known_modules: set[str]) -> str | None:
    matches = [
        module
        for module in known_modules
        if candidate == module or candidate.startswith(module + ".")
    ]
    if not matches:
        return None
    return max(matches, key=len)


def _normalize_import_from(
    node: ast.ImportFrom, current_package: str, known_modules: set[str]
) -> set[str]:
    if node.level == 0:
        base_parts = ()
    else:
        current_parts = current_package.split(".")
        if node.level == 1:
            base_parts = tuple(current_parts)
        else:
            base_parts = tuple(current_parts[: -(node.level - 1)])
    module_parts = tuple(node.module.split(".")) if node.module else ()
    base_module = ".".join(base_parts + module_parts)
    discovered: set[str] = set()
    if base_module.startswith("dagzoo"):
        normalized = _normalize_import(base_module, known_modules)
        if normalized is not None:
            discovered.add(normalized)
    for alias in node.names:
        if alias.name == "*":
            continue
        if base_module:
            child_candidate = f"{base_module}.{alias.name}"
        else:
            child_candidate = alias.name
        normalized_child = _normalize_import(child_candidate, known_modules)
        if normalized_child is not None:
            discovered.add(normalized_child)
    return discovered


def strongly_connected_components(graph: dict[str, set[str]]) -> list[tuple[str, ...]]:
    index = 0
    indices: dict[str, int] = {}
    lowlinks: dict[str, int] = {}
    stack: list[str] = []
    on_stack: set[str] = set()
    components: list[tuple[str, ...]] = []

    def visit(node: str) -> None:
        nonlocal index
        indices[node] = index
        lowlinks[node] = index
        index += 1
        stack.append(node)
        on_stack.add(node)

        for dependency in sorted(graph.get(node, ())):
            if dependency not in indices:
                visit(dependency)
                lowlinks[node] = min(lowlinks[node], lowlinks[dependency])
            elif dependency in on_stack:
                lowlinks[node] = min(lowlinks[node], indices[dependency])

        if lowlinks[node] != indices[node]:
            return

        component: list[str] = []
        while stack:
            current = stack.pop()
            on_stack.remove(current)
            component.append(current)
            if current == node:
                break
        components.append(tuple(sorted(component)))

    for node in sorted(graph):
        if node not in indices:
            visit(node)
    return sorted(components, key=lambda component: (len(component), component))


def condensation_graph(
    graph: dict[str, set[str]],
) -> tuple[list[tuple[str, ...]], dict[int, set[int]]]:
    components = strongly_connected_components(graph)
    index_by_node = {
        node: component_index
        for component_index, component in enumerate(components)
        for node in component
    }
    condensed: dict[int, set[int]] = {index: set() for index in range(len(components))}
    for node, dependencies in graph.items():
        source_index = index_by_node[node]
        for dependency in dependencies:
            target_index = index_by_node[dependency]
            if source_index != target_index:
                condensed[source_index].add(target_index)
    return components, condensed


def topological_order(graph: dict[int, set[int]]) -> list[int]:
    indegree = {node: 0 for node in graph}
    for dependencies in graph.values():
        for dependency in dependencies:
            indegree[dependency] = indegree.get(dependency, 0) + 1
    ready = sorted(node for node, degree in indegree.items() if degree == 0)
    order: list[int] = []
    while ready:
        node = ready.pop(0)
        order.append(node)
        for dependency in sorted(graph.get(node, ())):
            indegree[dependency] -= 1
            if indegree[dependency] == 0:
                ready.append(dependency)
                ready.sort()
    return order


def package_dag_lines(graph: ImportGraph) -> list[str]:
    package_graph = graph.package_graph()
    components, condensed = condensation_graph(package_graph)
    lines: list[str] = []
    for component_index in topological_order(condensed):
        component = components[component_index]
        label = ", ".join(f"`{name}`" for name in component)
        dependencies = [
            ", ".join(f"`{name}`" for name in components[target_index])
            for target_index in sorted(condensed[component_index])
        ]
        if dependencies:
            lines.append(f"- {label} depends on {', '.join(dependencies)}")
        else:
            lines.append(f"- {label} has no internal package dependencies")
    return lines


def hotspot_summaries(
    graph: ImportGraph, modules: tuple[str, ...] = HOTSPOT_MODULES
) -> dict[str, ModuleSummary]:
    return {module: graph.module_summary(module) for module in modules if module in graph.imports}


def render_dependency_map_markdown(graph: ImportGraph) -> str:
    lines = [
        "# Module Dependency Map",
        "",
        "This file is generated from imports under `src/dagzoo`.",
        "Run `./scripts/dev deps --write-docs` after changing internal module edges.",
        "",
        "## Package Dependency DAG",
        "",
        "The package graph below collapses strongly connected components so the result stays acyclic.",
        "",
        *package_dag_lines(graph),
        "",
        "## Change-Impact Hotspots",
        "",
        "The sections below list direct importers and full transitive downstream modules.",
        "Use them to predict which runtime paths are likely to move when a hot module changes.",
        "",
    ]
    for module, summary in hotspot_summaries(graph).items():
        lines.extend(
            [
                f"### `{module}`",
                "",
                f"- Path: `{graph.module_to_path(module)}`",
                f"- Imports: {_inline_list(summary.imports)}",
                f"- Direct downstream modules: {_inline_list(summary.direct_importers)}",
                f"- Transitive downstream modules: {_inline_list(summary.transitive_importers)}",
                f"- Downstream package areas: {_inline_list(summary.impacted_packages)}",
                "",
            ]
        )
    return "\n".join(lines).rstrip() + "\n"


def _inline_list(values: tuple[str, ...]) -> str:
    if not values:
        return "none"
    return ", ".join(f"`{value}`" for value in values)


def render_scope_text(graph: ImportGraph, scope: str) -> str:
    if scope == "package":
        return "\n".join(package_dag_lines(graph)) + "\n"
    if scope == "full":
        payload = {
            module: {
                "imports": list(graph.imports[module]),
                "imported_by": list(graph.reverse_imports.get(module, ())),
            }
            for module in graph.modules
        }
        return json.dumps(payload, indent=2) + "\n"
    return render_dependency_map_markdown(graph)


def render_scope_json(graph: ImportGraph, scope: str) -> str:
    if scope == "package":
        payload = {
            package: sorted(dependencies) for package, dependencies in graph.package_graph().items()
        }
        return json.dumps(payload, indent=2) + "\n"
    if scope == "full":
        payload = {
            module: {
                "imports": list(graph.imports[module]),
                "imported_by": list(graph.reverse_imports.get(module, ())),
            }
            for module in graph.modules
        }
        return json.dumps(payload, indent=2) + "\n"
    payload = {
        "package_dag": package_dag_lines(graph),
        "hotspots": {
            module: {
                "imports": list(summary.imports),
                "direct_downstream": list(summary.direct_importers),
                "transitive_downstream": list(summary.transitive_importers),
                "downstream_packages": list(summary.impacted_packages),
            }
            for module, summary in hotspot_summaries(graph).items()
        },
    }
    return json.dumps(payload, indent=2) + "\n"


def write_dependency_docs(graph: ImportGraph) -> str:
    content = render_dependency_map_markdown(graph)
    DOCS_DEP_MAP_PATH.write_text(content)
    return content


def dependency_docs_are_current(graph: ImportGraph) -> bool:
    if not DOCS_DEP_MAP_PATH.exists():
        return False
    return DOCS_DEP_MAP_PATH.read_text() == render_dependency_map_markdown(graph)



"""Tests for math/random_matrices.py."""

import pytest
import torch
from conftest import make_generator as _make_generator
from hypothesis import given, settings
from hypothesis import strategies as st

from dagzoo.math.random_matrices import sample_random_matrix
from dagzoo.rng import SEED32_MAX
from dagzoo.sampling.noise import NoiseSamplingSpec

_SEED32_STRATEGY = st.integers(min_value=0, max_value=SEED32_MAX)
_OUT_DIM_STRATEGY = st.integers(min_value=1, max_value=24)
_IN_DIM_STRATEGY = st.integers(min_value=1, max_value=24)
_KIND_STRATEGY = st.sampled_from(("gaussian", "weights", "singular_values", "kernel", "activation"))
_NORMALIZING_KIND_STRATEGY = st.sampled_from(("weights", "kernel", "activation"))
_NOISE_SIGMA_MULTIPLIER_STRATEGY = st.floats(
    min_value=0.125,
    max_value=4.0,
    allow_nan=False,
    allow_infinity=False,
    width=32,
)
_SCALE_STRATEGY = st.floats(
    min_value=0.125,
    max_value=4.0,
    allow_nan=False,
    allow_infinity=False,
    width=32,
)
_STUDENT_T_DF_STRATEGY = st.floats(
    min_value=2.25,
    max_value=12.0,
    allow_nan=False,
    allow_infinity=False,
    width=32,
)
_POSITIVE_WEIGHT_STRATEGY = st.floats(
    min_value=0.125,
    max_value=8.0,
    allow_nan=False,
    allow_infinity=False,
    width=32,
)
_MIXTURE_WEIGHTS_STRATEGY = st.dictionaries(
    keys=st.sampled_from(("gaussian", "laplace", "student_t")),
    values=_POSITIVE_WEIGHT_STRATEGY,
    min_size=1,
    max_size=3,
)


@st.composite
def _noise_spec_strategy(draw: st.DrawFn) -> NoiseSamplingSpec | None:
    if draw(st.booleans()):
        return None

    family = draw(st.sampled_from(("gaussian", "laplace", "student_t", "mixture")))
    scale = draw(_SCALE_STRATEGY)
    if family == "student_t":
        return NoiseSamplingSpec(
            family=family,
            scale=scale,
            student_t_df=draw(_STUDENT_T_DF_STRATEGY),
        )
    if family == "mixture":
        return NoiseSamplingSpec(
            family=family,
            scale=scale,
            student_t_df=draw(_STUDENT_T_DF_STRATEGY),
            mixture_weights=draw(_MIXTURE_WEIGHTS_STRATEGY),
        )
    return NoiseSamplingSpec(family=family, scale=scale)


def _assert_valid_matrix(matrix: torch.Tensor, *, out_dim: int, in_dim: int) -> None:
    assert matrix.shape == (out_dim, in_dim)
    assert torch.all(torch.isfinite(matrix))
    norms = torch.linalg.norm(matrix, dim=1)
    torch.testing.assert_close(
        norms,
        torch.ones(out_dim, dtype=matrix.dtype, device=matrix.device),
        atol=1e-4,
        rtol=1e-4,
    )


def test_output_shape() -> None:
    g = _make_generator(0)
    m = sample_random_matrix(5, 3, g, "cpu")
    assert m.shape == (5, 3)


def test_rows_unit_normalized() -> None:
    g = _make_generator(1)
    m = sample_random_matrix(4, 6, g, "cpu")
    norms = torch.linalg.norm(m, dim=1)
    torch.testing.assert_close(norms, torch.ones(4), atol=1e-4, rtol=1e-4)


def test_deterministic() -> None:
    a = sample_random_matrix(3, 4, _make_generator(42), "cpu")
    b = sample_random_matrix(3, 4, _make_generator(42), "cpu")
    torch.testing.assert_close(a, b)


def test_kernel_single_column_rows_are_unit_normalized() -> None:
    matrix = sample_random_matrix(6, 1, _make_generator(0), "cpu", kind="kernel")
    norms = torch.linalg.norm(matrix, dim=1)
    torch.testing.assert_close(norms, torch.ones(6), atol=1e-4, rtol=1e-4)


@pytest.mark.parametrize("kind", ["gaussian", "weights", "singular_values", "kernel", "activation"])
def test_each_kind(kind: str) -> None:
    g = _make_generator(10)
    m = sample_random_matrix(4, 5, g, "cpu", kind=kind)
    assert m.shape == (4, 5)
    assert torch.all(torch.isfinite(m))


def test_invalid_kind_raises() -> None:
    with pytest.raises(ValueError, match="Unknown matrix kind"):
        sample_random_matrix(2, 2, _make_generator(0), "cpu", kind="bogus")


@settings(max_examples=60, deadline=None)
@given(
    out_dim=_OUT_DIM_STRATEGY,
    in_dim=_IN_DIM_STRATEGY,
    seed=_SEED32_STRATEGY,
    kind=_KIND_STRATEGY,
    noise_sigma_multiplier=_NOISE_SIGMA_MULTIPLIER_STRATEGY,
    noise_spec=_noise_spec_strategy(),
)
def test_random_matrix_hypothesis_supported_kinds_are_finite_normalized_and_deterministic(
    out_dim: int,
    in_dim: int,
    seed: int,
    kind: str,
    noise_sigma_multiplier: float,
    noise_spec: NoiseSamplingSpec | None,
) -> None:
    first = sample_random_matrix(
        out_dim,
        in_dim,
        _make_generator(seed),
        "cpu",
        kind=kind,
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )
    second = sample_random_matrix(
        out_dim,
        in_dim,
        _make_generator(seed),
        "cpu",
        kind=kind,
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )

    _assert_valid_matrix(first, out_dim=out_dim, in_dim=in_dim)
    torch.testing.assert_close(first, second)


@settings(max_examples=60, deadline=None)
@given(
    out_dim=_OUT_DIM_STRATEGY,
    in_dim=_IN_DIM_STRATEGY,
    seed=_SEED32_STRATEGY,
    noise_sigma_multiplier=_NOISE_SIGMA_MULTIPLIER_STRATEGY,
    noise_spec=_noise_spec_strategy(),
)
def test_random_matrix_hypothesis_kind_none_is_deterministic_and_valid(
    out_dim: int,
    in_dim: int,
    seed: int,
    noise_sigma_multiplier: float,
    noise_spec: NoiseSamplingSpec | None,
) -> None:
    first = sample_random_matrix(
        out_dim,
        in_dim,
        _make_generator(seed),
        "cpu",
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )
    second = sample_random_matrix(
        out_dim,
        in_dim,
        _make_generator(seed),
        "cpu",
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )

    _assert_valid_matrix(first, out_dim=out_dim, in_dim=in_dim)
    torch.testing.assert_close(first, second)


@settings(max_examples=60, deadline=None)
@given(
    out_dim=_OUT_DIM_STRATEGY,
    in_dim=_IN_DIM_STRATEGY,
    seed=_SEED32_STRATEGY,
    kind=_NORMALIZING_KIND_STRATEGY,
    noise_sigma_multiplier=_NOISE_SIGMA_MULTIPLIER_STRATEGY,
    noise_spec=_noise_spec_strategy(),
)
def test_random_matrix_hypothesis_transformed_kinds_preserve_row_normalization(
    out_dim: int,
    in_dim: int,
    seed: int,
    kind: str,
    noise_sigma_multiplier: float,
    noise_spec: NoiseSamplingSpec | None,
) -> None:
    matrix = sample_random_matrix(
        out_dim,
        in_dim,
        _make_generator(seed),
        "cpu",
        kind=kind,
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )

    _assert_valid_matrix(matrix, out_dim=out_dim, in_dim=in_dim)



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import sys
from pathlib import Path

SCRIPT_DIR = Path(__file__).resolve().parent
if str(SCRIPT_DIR) not in sys.path:
    sys.path.insert(0, str(SCRIPT_DIR))

from devlib.common import DOCS_DEP_MAP_PATH, DevToolError, repo_relative  # noqa: E402
from devlib.contract import evaluate_release_contract, render_contract_result  # noqa: E402
from devlib.deps import (  # noqa: E402
    build_import_graph,
    dependency_docs_are_current,
    render_scope_json,
    render_scope_text,
    write_dependency_docs,
)
from devlib.doctor import doctor_passed, render_doctor_results, run_doctor  # noqa: E402
from devlib.impact import (  # noqa: E402
    build_impact_report,
    detect_changed_files,
)
from devlib.impact import render_json as render_impact_json  # noqa: E402
from devlib.impact import render_text as render_impact_text  # noqa: E402
from devlib.verify import build_verify_plan, execute_verify_plan  # noqa: E402


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="./scripts/dev")
    subparsers = parser.add_subparsers(dest="command", required=True)

    doctor_parser = subparsers.add_parser("doctor")
    doctor_parser.add_argument("mode", choices=("code", "docs", "all"), nargs="?", default="all")

    deps_parser = subparsers.add_parser("deps")
    deps_parser.add_argument("--scope", choices=("package", "hybrid", "full"), default="hybrid")
    deps_parser.add_argument("--format", choices=("text", "json"), default="text")
    deps_parser.add_argument("--write-docs", action="store_true")
    deps_parser.add_argument("--check", action="store_true")

    impact_parser = subparsers.add_parser("impact")
    _add_change_source_args(impact_parser)
    impact_parser.add_argument("--format", choices=("text", "json"), default="text")

    contract_parser = subparsers.add_parser("contract")
    _add_change_source_args(contract_parser)
    contract_parser.add_argument("--strict", action="store_true")

    verify_parser = subparsers.add_parser("verify")
    verify_parser.add_argument(
        "mode", choices=("quick", "code", "docs", "bench", "full", "affected")
    )
    _add_change_source_args(verify_parser)
    verify_parser.add_argument("--dry-run", action="store_true")
    verify_parser.add_argument("--incremental", action="store_true")
    verify_parser.add_argument("--parallel", action="store_true")

    return parser


def _add_change_source_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--source", choices=("working-tree", "staged", "base"), default="working-tree"
    )
    parser.add_argument("--base")
    parser.add_argument("--files", nargs="*")


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    try:
        if args.command == "doctor":
            results = run_doctor(args.mode)
            print(render_doctor_results(results), end="")
            return 0 if doctor_passed(results) else 1

        if args.command == "deps":
            graph = build_import_graph()
            if args.write_docs:
                write_dependency_docs(graph)
                print(f"wrote {repo_relative(DOCS_DEP_MAP_PATH)}")
            if args.check:
                if not dependency_docs_are_current(graph):
                    print(
                        "dependency map docs are stale; run `./scripts/dev deps --write-docs`.",
                        file=sys.stderr,
                    )
                    return 1
                if not args.write_docs:
                    print("dependency map docs are current.")
                return 0
            if args.write_docs:
                return 0
            rendered = (
                render_scope_json(graph, args.scope)
                if args.format == "json"
                else render_scope_text(graph, args.scope)
            )
            print(rendered, end="")
            return 0

        if args.command == "impact":
            changed_files = detect_changed_files(
                source=args.source,
                base=args.base,
                files=args.files,
            )
            report = build_impact_report(changed_files)
            rendered = (
                render_impact_json(report) if args.format == "json" else render_impact_text(report)
            )
            print(rendered, end="")
            return 0

        if args.command == "contract":
            changed_files = detect_changed_files(
                source=args.source,
                base=args.base,
                files=args.files,
            )
            report = build_impact_report(changed_files)
            result = evaluate_release_contract(report, strict=args.strict)
            print(render_contract_result(result), end="")
            return 0 if result.ok else 1

        if args.command == "verify":
            plan = build_verify_plan(
                mode=args.mode,
                source=args.source,
                base=args.base,
                files=args.files,
                incremental=args.incremental,
                parallel=args.parallel,
            )
            print(execute_verify_plan(plan, dry_run=args.dry_run), end="")
            return 0
    except DevToolError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    except RuntimeError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    parser.error("unreachable")
    return 2


if __name__ == "__main__":
    raise SystemExit(main())



import json
from pathlib import Path

import yaml
from conftest import load_repo_config, write_config

from dagzoo.bench.constants import SMOKE_N_TEST_CAP, SMOKE_N_TRAIN_CAP
from dagzoo.cli import (
    DEVICE_CHOICES,
    HARDWARE_POLICY_CHOICES,
    MISSINGNESS_MECHANISM_CLI_CHOICES,
    _print_preset_result_line,
    main,
)


def test_cli_package_reexports_parser_choice_constants() -> None:
    assert DEVICE_CHOICES == ("auto", "cpu", "cuda", "mps")
    assert "none" in HARDWARE_POLICY_CHOICES
    assert tuple(MISSINGNESS_MECHANISM_CLI_CHOICES) == ("none", "mcar", "mar", "mnar")


def test_benchmark_cli_writes_json(tmp_path) -> None:
    out = tmp_path / "summary.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    assert out.exists()
    payload = json.loads(out.read_text(encoding="utf-8"))
    assert payload["suite"] == "smoke"
    assert len(payload["preset_results"]) == 1
    profile = payload["preset_results"][0]
    assert profile["generation_mode"] == "fixed_batched"
    assert float(profile["generation_datasets_per_minute"]) >= 0.0
    assert float(profile["write_datasets_per_minute"]) >= 0.0
    assert profile["filter_datasets_per_minute"] is None
    assert profile["filter_accepted_datasets_per_minute"] is None
    assert "accepted_datasets_measured" not in profile
    assert profile["filter_accepted_datasets_measured"] == 0
    assert profile["filter_rejected_datasets_measured"] == 0
    assert profile["filter_acceptance_rate_dataset_level"] is None
    assert profile["filter_rejection_rate_dataset_level"] is None
    assert profile["filter_rejection_rate_attempt_level"] is None
    assert profile["filter_retry_dataset_rate"] is None
    lineage_guardrails = profile["lineage_guardrails"]
    assert isinstance(lineage_guardrails, dict)
    assert isinstance(lineage_guardrails["enabled"], bool)
    if lineage_guardrails["enabled"]:
        assert lineage_guardrails["status"] in {"pass", "warn", "fail"}


def test_benchmark_cli_realizes_dataset_rows_once_per_preset(tmp_path) -> None:
    cfg = load_repo_config()
    cfg.dataset.rows = "400..60000"  # type: ignore[assignment]
    config_path = write_config(tmp_path, cfg, "rows_config.yaml")
    out_dir = tmp_path / "rows_benchmark"

    code = main(
        [
            "benchmark",
            "--config",
            str(config_path),
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
        ]
    )

    assert code == 0
    payload = json.loads((out_dir / "summary.json").read_text(encoding="utf-8"))
    result = payload["preset_results"][0]
    effective_config = result["effective_config"]
    assert int(result["dataset_rows_total"]) <= int(SMOKE_N_TRAIN_CAP + SMOKE_N_TEST_CAP)
    assert effective_config["dataset"]["rows"] is None
    trace_files = sorted((out_dir / "effective_configs").glob("*_trace.yaml"))
    assert trace_files
    trace_payload = yaml.safe_load(trace_files[0].read_text(encoding="utf-8"))
    assert any(
        isinstance(item, dict) and item.get("source") == "benchmark.smoke_rows_cap"
        for item in trace_payload
    )


def test_benchmark_cli_writes_effective_config_trace_artifacts(tmp_path) -> None:
    out_dir = tmp_path / "bench_results"
    code = main(
        [
            "benchmark",
            "--preset",
            "cpu",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
        ]
    )
    assert code == 0
    trace_files = sorted((out_dir / "effective_configs").glob("*_trace.yaml"))
    assert trace_files
    trace_payload = yaml.safe_load(trace_files[0].read_text(encoding="utf-8"))
    assert isinstance(trace_payload, list)
    assert any(
        isinstance(item, dict) and item.get("source") == "benchmark.suite_smoke_caps"
        for item in trace_payload
    )


def test_benchmark_cli_builtin_cpu_reports_fixed_batched_generation_mode(tmp_path) -> None:
    out = tmp_path / "summary.json"
    code = main(
        [
            "benchmark",
            "--preset",
            "cpu",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    preset_results = payload["preset_results"]
    assert [result["preset_key"] for result in preset_results] == [
        "cpu_rows1024",
        "cpu_rows4096",
        "cpu_rows8192",
    ]
    assert [result["dataset_rows_total"] for result in preset_results] == [1024, 4096, 8192]
    assert all(result["generation_mode"] == "fixed_batched" for result in preset_results)


def test_benchmark_cli_filter_smoke_config_reports_accepted_corpus_throughput(tmp_path) -> None:
    out_dir = tmp_path / "filter_smoke_results"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/preset_filter_benchmark_smoke.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
        ]
    )
    assert code == 0
    payload = json.loads((out_dir / "summary.json").read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    assert profile["filter_datasets_per_minute"] is not None
    assert float(profile["filter_accepted_datasets_per_minute"]) > 0.0
    assert int(profile["filter_accepted_datasets_measured"]) > 0
    assert int(profile["filter_rejected_datasets_measured"]) > 0
    assert 0.0 < float(profile["filter_acceptance_rate_dataset_level"]) < 1.0
    assert 0.0 < float(profile["filter_rejection_rate_dataset_level"]) < 1.0


def test_benchmark_cli_fail_on_regression(tmp_path) -> None:
    baseline_path = tmp_path / "baseline.json"
    baseline_payload = {
        "version": 1,
        "suite": "standard",
        "metrics": ["datasets_per_minute"],
        "presets": {
            "medium_cuda": {"datasets_per_minute": 1.0e9},
        },
    }
    baseline_path.write_text(json.dumps(baseline_payload), encoding="utf-8")

    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--baseline",
            str(baseline_path),
            "--warn-threshold-pct",
            "1",
            "--fail-threshold-pct",
            "2",
            "--fail-on-regression",
        ]
    )
    assert code == 1


def test_benchmark_cli_fail_on_regression_for_filter_accepted_throughput(
    tmp_path, monkeypatch, capsys
) -> None:
    baseline_path = tmp_path / "baseline.json"
    baseline_payload = {
        "version": 1,
        "suite": "smoke",
        "metrics": ["filter_accepted_datasets_per_minute"],
        "presets": {
            "custom": {"filter_accepted_datasets_per_minute": 100.0},
        },
    }
    baseline_path.write_text(json.dumps(baseline_payload), encoding="utf-8")

    summary = {
        "suite": "smoke",
        "preset_results": [
            {
                "preset_key": "custom",
                "filter_accepted_datasets_per_minute": 70.0,
                "datasets_per_minute": 1000.0,
                "latency_p95_ms": 1.0,
            }
        ],
        "regression": {
            "status": "fail",
            "issues": [
                {
                    "preset": "custom",
                    "metric": "filter_accepted_datasets_per_minute",
                    "current": 70.0,
                    "baseline": 100.0,
                    "degradation_pct": 30.0,
                    "severity": "fail",
                }
            ],
            "hard_fail": True,
        },
    }

    calls: list[dict[str, object]] = []

    def _fake_run_benchmark_suite(*args, **kwargs):
        calls.append(kwargs)
        return summary

    monkeypatch.setattr("dagzoo.cli.run_benchmark_suite", _fake_run_benchmark_suite)

    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--baseline",
            str(baseline_path),
            "--warn-threshold-pct",
            "1",
            "--fail-threshold-pct",
            "2",
            "--fail-on-regression",
        ]
    )

    assert code == 1
    assert calls == [
        {
            "suite": "smoke",
            "warn_threshold_pct": 1.0,
            "fail_threshold_pct": 2.0,
            "baseline_payload": baseline_payload,
            "num_datasets_override": 2,
            "warmup_override": 0,
            "collect_memory": False,
            "collect_reproducibility": False,
            "collect_diagnostics": False,
            "diagnostics_root_dir": None,
            "fail_on_regression": True,
            "hardware_policy": "none",
        }
    ]
    captured = capsys.readouterr()
    assert "Regression status=fail issues=1" in captured.out
    assert "filter_accepted/min=70.00" in captured.out


def test_benchmark_cli_diagnostics_emits_artifacts(tmp_path) -> None:
    out_dir = tmp_path / "bench_results"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--diagnostics",
            "--out-dir",
            str(out_dir),
        ]
    )
    assert code == 0

    summary_path = out_dir / "summary.json"
    assert summary_path.exists()
    payload = json.loads(summary_path.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    assert profile["diagnostics_enabled"] is True

    artifacts = profile["diagnostics_artifacts"]
    assert isinstance(artifacts, dict)
    assert isinstance(artifacts["json"], str)
    assert isinstance(artifacts["markdown"], str)
    assert Path(artifacts["json"]).exists()
    assert Path(artifacts["markdown"]).exists()


def test_benchmark_cli_json_out_diagnostics_does_not_emit_default_summary_artifacts(
    tmp_path, monkeypatch
) -> None:
    repo_root = Path(__file__).resolve().parents[1]
    config_path = repo_root / "configs" / "default.yaml"
    json_out = tmp_path / "summary.json"
    diagnostics_out = tmp_path / "diag_out"
    monkeypatch.chdir(tmp_path)

    code = main(
        [
            "benchmark",
            "--config",
            str(config_path),
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--diagnostics",
            "--json-out",
            str(json_out),
            "--diagnostics-out-dir",
            str(diagnostics_out),
        ]
    )
    assert code == 0
    assert json_out.exists()
    assert not (tmp_path / "benchmarks").exists()


def test_benchmark_cli_diagnostics_pointers_resolve_when_roots_differ(tmp_path) -> None:
    out_dir = tmp_path / "bench_results"
    diagnostics_out = tmp_path / "diag_out"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--diagnostics",
            "--out-dir",
            str(out_dir),
            "--diagnostics-out-dir",
            str(diagnostics_out),
        ]
    )
    assert code == 0

    summary_path = out_dir / "summary.json"
    assert summary_path.exists()
    payload = json.loads(summary_path.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    artifacts = profile["diagnostics_artifacts"]
    assert isinstance(artifacts, dict)
    json_path = Path(artifacts["json"])
    md_path = Path(artifacts["markdown"])
    assert json_path.is_absolute()
    assert md_path.is_absolute()
    assert json_path.exists()
    assert md_path.exists()
    assert json_path.resolve().is_relative_to(diagnostics_out.resolve())
    assert md_path.resolve().is_relative_to(diagnostics_out.resolve())


def test_benchmark_cli_missingness_guardrails_are_emitted(tmp_path) -> None:
    out = tmp_path / "summary_missingness.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/preset_missingness_mcar.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    guardrails = profile["missingness_guardrails"]
    assert guardrails["enabled"] is True
    assert guardrails["status"] in {"pass", "warn", "fail"}
    lineage_guardrails = profile["lineage_guardrails"]
    assert isinstance(lineage_guardrails, dict)
    assert isinstance(lineage_guardrails["enabled"], bool)
    if lineage_guardrails["enabled"]:
        assert lineage_guardrails["status"] in {"pass", "warn", "fail"}


def test_benchmark_cli_shift_guardrails_are_emitted(tmp_path) -> None:
    out = tmp_path / "summary_shift.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/preset_shift_benchmark_smoke.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    guardrails = profile["shift_guardrails"]
    assert guardrails["enabled"] is True
    assert guardrails["status"] in {"pass", "warn", "fail"}
    assert "directional_checks" in guardrails
    assert "runtime_gating_enabled" in guardrails


def test_benchmark_cli_noise_guardrails_are_emitted(tmp_path) -> None:
    out = tmp_path / "summary_noise.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/preset_noise_benchmark_smoke.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    guardrails = profile["noise_guardrails"]
    assert guardrails["enabled"] is True
    assert guardrails["status"] in {"pass", "warn", "fail"}
    assert "metadata_coverage_rate" in guardrails
    assert "runtime_gating_enabled" in guardrails


def test_benchmark_cli_many_class_smoke_preset_emits_runtime_metrics(tmp_path) -> None:
    out = tmp_path / "summary_many_class_smoke.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/preset_many_class_benchmark_smoke.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    assert float(profile["datasets_per_minute"]) > 0.0
    assert float(profile["latency_p95_ms"]) >= 0.0
    regression = payload["regression"]
    assert regression["status"] in {"pass", "warn", "fail"}
    guardrails = profile["lineage_guardrails"]
    assert isinstance(guardrails, dict)
    assert isinstance(guardrails["enabled"], bool)


def test_print_preset_result_line_includes_shift_status(capsys) -> None:
    _print_preset_result_line(
        {
            "preset_key": "shift_smoke",
            "device": "cpu",
            "hardware_backend": "cpu",
            "datasets_per_minute": 123.0,
            "latency_p95_ms": 4.2,
            "shift_guardrails": {"enabled": True, "status": "pass"},
        }
    )
    output = capsys.readouterr().out
    assert "shift=pass" in output


def test_print_preset_result_line_includes_noise_status(capsys) -> None:
    _print_preset_result_line(
        {
            "preset_key": "noise_smoke",
            "device": "cpu",
            "hardware_backend": "cpu",
            "datasets_per_minute": 123.0,
            "latency_p95_ms": 4.2,
            "noise_guardrails": {"enabled": True, "status": "warn"},
        }
    )
    output = capsys.readouterr().out
    assert "noise=warn" in output


def test_print_preset_result_line_includes_stage_and_filter_rejection_metrics(capsys) -> None:
    _print_preset_result_line(
        {
            "preset_key": "throughput_smoke",
            "device": "cpu",
            "hardware_backend": "cpu",
            "datasets_per_minute": 123.0,
            "generation_datasets_per_minute": 124.0,
            "write_datasets_per_minute": 80.0,
            "filter_datasets_per_minute": 60.0,
            "filter_accepted_datasets_per_minute": 45.0,
            "filter_acceptance_rate_dataset_level": 0.75,
            "filter_rejection_rate_dataset_level": 0.25,
            "filter_rejection_rate_attempt_level": 0.125,
            "filter_retry_dataset_rate": 0.25,
            "latency_p95_ms": 4.2,
        }
    )
    output = capsys.readouterr().out
    assert "gen/min=124.00" in output
    assert "write/min=80.00" in output
    assert "filter/min=60.00" in output
    assert "filter_accepted/min=45.00" in output
    assert "filter_reject_attempt_pct=12.50" in output
    assert "filter_accept_dataset_pct=75.00" in output
    assert "filter_reject_dataset_pct=25.00" in output
    assert "filter_retry_dataset_pct=25.00" in output


def test_print_preset_result_line_handles_missing_latency(capsys) -> None:
    _print_preset_result_line(
        {
            "preset_key": "parallel_cpu",
            "device": "cpu",
            "hardware_backend": "cpu",
            "datasets_per_minute": 123.0,
            "generation_datasets_per_minute": 124.0,
            "write_datasets_per_minute": 80.0,
            "latency_p95_ms": None,
        }
    )
    output = capsys.readouterr().out
    assert "latency_p95_ms=-" in output



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path

_PYPROJECT_VERSION_RE = re.compile(r'^version = "([^"]+)"$', re.MULTILINE)


@dataclass(frozen=True, order=True)
class SemVer:
    major: int
    minor: int
    patch: int

    @classmethod
    def parse(cls, value: str) -> "SemVer":
        match = re.fullmatch(r"(\d+)\.(\d+)\.(\d+)", str(value).strip())
        if match is None:
            raise ValueError(f"Unsupported version format: {value!r}")
        return cls(*(int(part) for part in match.groups()))

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


def read_pyproject_version_text(text: str, *, source: str) -> str:
    match = _PYPROJECT_VERSION_RE.search(text)
    if match is None:
        raise ValueError(f"Could not parse version from {source!r}")
    return match.group(1)


def read_pyproject_version(pyproject_path: str) -> str:
    return read_pyproject_version_text(
        Path(pyproject_path).read_text(encoding="utf-8"),
        source=pyproject_path,
    )


def allowed_release_successors(previous: SemVer) -> set[SemVer]:
    return {
        SemVer(previous.major, previous.minor, previous.patch + 1),
        SemVer(previous.major, previous.minor + 1, 0),
        SemVer(previous.major + 1, 0, 0),
    }


def allowed_pr_successors(previous: SemVer) -> set[SemVer]:
    return {
        SemVer(previous.major, previous.minor, previous.patch + 1),
        SemVer(previous.major, previous.minor + 1, 0),
    }



from __future__ import annotations

import importlib.util
import sys
from functools import lru_cache
from pathlib import Path

import pytest
import torch
import yaml

from dagzoo.config import GeneratorConfig, clone_generator_config
from dagzoo.hardware import HardwareInfo
from dagzoo.rng import KeyedRng, keyed_rng_from_generator

REPO_ROOT = Path(__file__).resolve().parents[1]


def make_generator(seed: int = 42) -> torch.Generator:
    """Create a seeded torch Generator on CPU for deterministic tests."""
    g = torch.Generator(device="cpu")
    g.manual_seed(seed)
    return g


def make_keyed_rng(generator: torch.Generator, *components: str | int) -> KeyedRng:
    """Consume one ambient draw and derive the same keyed root used by helpers."""

    return keyed_rng_from_generator(generator, *components)


@lru_cache(maxsize=None)
def _cached_repo_config(resource_name: str) -> GeneratorConfig:
    return GeneratorConfig.from_yaml(REPO_ROOT / "configs" / resource_name)


def load_repo_config(resource_name: str = "default.yaml") -> GeneratorConfig:
    """Load one repo config and return a fresh mutable copy for the caller."""

    return clone_generator_config(_cached_repo_config(resource_name), revalidate=False)


def write_yaml(tmp_path: Path, name: str, payload: object) -> Path:
    """Write one YAML payload under ``tmp_path`` and return the created path."""

    path = tmp_path / name
    path.write_text(yaml.safe_dump(payload, sort_keys=False), encoding="utf-8")
    return path


def write_config(tmp_path: Path, config: GeneratorConfig, name: str = "config.yaml") -> Path:
    """Serialize one ``GeneratorConfig`` under ``tmp_path`` and return the path."""

    return write_yaml(tmp_path, name, config.to_dict())


def load_script_module(module_name: str, rel_path: str):
    """Load one repo script module from a repo-relative path."""

    script_path = REPO_ROOT / rel_path
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def _build_mock_hardware(tier: str) -> HardwareInfo:
    normalized_tier = str(tier).strip().lower()
    if normalized_tier == "cpu":
        return HardwareInfo(
            backend="cpu",
            requested_device="cpu",
            device_name="cpu",
            total_memory_gb=None,
            peak_flops=float("inf"),
            tier="cpu",
        )
    if normalized_tier == "cuda_desktop":
        return HardwareInfo(
            backend="cuda",
            requested_device="cuda",
            device_name="NVIDIA RTX 4090",
            total_memory_gb=24.0,
            peak_flops=165.2e12,
            tier="cuda_desktop",
        )
    if normalized_tier == "cuda_datacenter":
        return HardwareInfo(
            backend="cuda",
            requested_device="cuda",
            device_name="NVIDIA A100 80GB",
            total_memory_gb=80.0,
            peak_flops=312e12,
            tier="cuda_datacenter",
        )
    if normalized_tier == "cuda_h100":
        return HardwareInfo(
            backend="cuda",
            requested_device="cuda",
            device_name="NVIDIA H100 SXM",
            total_memory_gb=80.0,
            peak_flops=989e12,
            tier="cuda_h100",
        )
    raise ValueError(f"Unsupported mock hardware tier {tier!r}.")


@pytest.fixture
def hardware_info_factory():
    """Return one helper that builds stable mock hardware profiles by tier."""

    return _build_mock_hardware


@pytest.fixture
def patch_detect_hardware(monkeypatch: pytest.MonkeyPatch, hardware_info_factory):
    """Patch one or more detect_hardware call sites to one stable mock tier."""

    def _patch(tier: str, *targets: str) -> HardwareInfo:
        hw = hardware_info_factory(tier)
        for target in targets:
            monkeypatch.setattr(target, lambda _requested_device, hw=hw: hw)
        return hw

    return _patch



"""Typed payloads for dataset-level diagnostics metrics."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class DatasetMetrics:
    """Stable dataset-level metrics payload used by diagnostics extractors."""

    task: str
    n_rows: int
    n_features: int
    n_classes: int | None
    n_categorical_features: int
    categorical_ratio: float
    graph_edge_density: float | None
    shift_enabled: float
    shift_graph_scale: float
    shift_mechanism_scale: float
    shift_variance_scale: float
    shift_edge_odds_multiplier: float
    shift_mechanism_nonlinear_mass: float
    shift_noise_variance_multiplier: float

    linearity_proxy: float | None
    nonlinearity_proxy: float | None
    wins_ratio_proxy: float | None

    pearson_abs_mean: float | None
    pearson_abs_max: float | None
    spearman_abs_mean: float | None
    spearman_abs_max: float | None

    class_entropy: float | None
    majority_minority_ratio: float | None
    snr_proxy_db: float | None

    cat_cardinality_min: int | None
    cat_cardinality_mean: float | None
    cat_cardinality_max: int | None



import pytest
import torch
from conftest import make_generator as _make_generator

import dagzoo.functions.random_functions as random_functions_mod
from dagzoo.core.execution_semantics import sample_function_family
from dagzoo.core.fixed_layout.plan_types import GaussianMatrixPlan, LinearFunctionPlan
from dagzoo.functions.random_functions import (
    MechanismFamily,
    _sample_function_family,
    apply_random_function,
)
from dagzoo.rng import KeyedRng


def test_tree_family_survives_nan_feature() -> None:
    g = _make_generator(42)
    x = torch.randn(64, 4, generator=g)
    x[:, 2] = float("nan")

    y = apply_random_function(x, _make_generator(42), out_dim=2, function_type="tree")
    assert y.shape == (64, 2)
    assert torch.all(torch.isfinite(y))

    x = torch.ones(32, 3)
    y = apply_random_function(x, _make_generator(7), out_dim=1, function_type="tree")
    assert y.shape == (32, 1)
    assert torch.all(torch.isfinite(y))


def test_output_shape() -> None:
    g = _make_generator(0)
    x = torch.randn(64, 4, generator=g)
    y = apply_random_function(x, g, out_dim=5)
    assert y.shape == (64, 5)


def test_deterministic() -> None:
    x = torch.randn(32, 4)
    y1 = apply_random_function(x.clone(), _make_generator(0), out_dim=3)
    y2 = apply_random_function(x.clone(), _make_generator(0), out_dim=3)
    torch.testing.assert_close(y1, y2)


def test_deterministic_with_shift_tilt_and_noise_multiplier() -> None:
    x = torch.randn(32, 4)
    y1 = apply_random_function(
        x.clone(),
        _make_generator(8),
        out_dim=3,
        mechanism_logit_tilt=0.7,
        noise_sigma_multiplier=1.35,
    )
    y2 = apply_random_function(
        x.clone(),
        _make_generator(8),
        out_dim=3,
        mechanism_logit_tilt=0.7,
        noise_sigma_multiplier=1.35,
    )
    torch.testing.assert_close(y1, y2)


def test_apply_random_function_output_depends_on_generator_root_nonce(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    roots = iter(
        [
            KeyedRng(seed=17, _ambient_nonce=(101, 102, 103)),
            KeyedRng(seed=17, _ambient_nonce=(201, 202, 203)),
        ]
    )
    monkeypatch.setattr(
        random_functions_mod,
        "keyed_rng_from_generator",
        lambda *_args, **_kwargs: next(roots),
    )
    monkeypatch.setattr(
        random_functions_mod,
        "sample_function_plan_for_family",
        lambda *_args, **_kwargs: LinearFunctionPlan(matrix=GaussianMatrixPlan()),
    )

    x = torch.randn(32, 4, generator=_make_generator(11))
    first = apply_random_function(x.clone(), _make_generator(12), out_dim=3, function_type="linear")
    second = apply_random_function(
        x.clone(),
        _make_generator(12),
        out_dim=3,
        function_type="linear",
    )

    assert not torch.equal(first, second)


@pytest.mark.parametrize(
    "family",
    ["linear", "quadratic", "discretization", "gp", "em", "product", "piecewise"],
)
def test_non_finite_inputs_are_sanitized(family: MechanismFamily) -> None:
    x = torch.tensor(
        [
            [0.0, float("nan"), 1.0],
            [float("inf"), -1.0, 2.0],
            [-float("inf"), 0.5, 3.0],
            [1.5, -2.0, 4.0],
            [2.0, 3.0, -5.0],
            [-3.0, 1.0, 6.0],
        ],
        dtype=torch.float32,
    )

    y = apply_random_function(x, _make_generator(9), out_dim=2, function_type=family)
    assert y.shape == (6, 2)
    assert torch.all(torch.isfinite(y))


@pytest.mark.parametrize(
    "family",
    ["nn", "tree", "discretization", "gp", "linear", "quadratic", "em", "product", "piecewise"],
)
def test_each_family(family: MechanismFamily) -> None:
    g = _make_generator(10)
    x = torch.randn(64, 4, generator=g)
    y = apply_random_function(x, _make_generator(10), out_dim=3, function_type=family)
    assert y.shape == (64, 3)
    assert torch.all(torch.isfinite(y))


@pytest.mark.parametrize("family", ["discretization", "quadratic"])
def test_explicit_family_is_deterministic(family: MechanismFamily) -> None:
    x = torch.randn(64, 4, generator=_make_generator(14))
    y1 = apply_random_function(
        x.clone(),
        _make_generator(15),
        out_dim=3,
        function_type=family,
    )
    y2 = apply_random_function(
        x.clone(),
        _make_generator(15),
        out_dim=3,
        function_type=family,
    )
    torch.testing.assert_close(y1, y2)


def test_invalid_type_raises() -> None:
    g = _make_generator()
    x = torch.randn(32, 4, generator=g)
    with pytest.raises(ValueError, match="Unknown random function family"):
        apply_random_function(x, g, out_dim=2, function_type="bogus")  # type: ignore[arg-type]


def test_sampled_family_respects_family_mix_allowlist() -> None:
    g = _make_generator(3)
    mix = {"linear": 1.0}
    for _ in range(16):
        sampled = sample_function_family(
            g,
            mechanism_logit_tilt=0.9,
            function_family_mix=mix,  # type: ignore[arg-type]
        )
        assert sampled == "linear"


def test_sampled_family_skips_zero_probability_families_on_zero_draw(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    g = _make_generator(33)
    mix = {"linear": 1.0}
    monkeypatch.setattr(torch, "rand", lambda *args, **kwargs: torch.tensor([0.0]))
    sampled = sample_function_family(
        g,
        mechanism_logit_tilt=0.9,
        function_family_mix=mix,  # type: ignore[arg-type]
    )
    assert sampled == "linear"


def test_disallowed_explicit_family_raises_when_mix_is_present() -> None:
    g = _make_generator(4)
    x = torch.randn(32, 4, generator=g)
    with pytest.raises(ValueError, match="is not enabled by mechanism.function_family_mix"):
        apply_random_function(
            x,
            g,
            out_dim=2,
            function_type="gp",
            function_family_mix={"linear": 1.0},  # type: ignore[arg-type]
        )


def test_product_family_without_enabled_component_raises_when_mix_is_present() -> None:
    g = _make_generator(5)
    x = torch.randn(32, 4, generator=g)
    with pytest.raises(ValueError, match="enables 'product' but disables all product component"):
        apply_random_function(
            x,
            g,
            out_dim=2,
            function_type="product",
            function_family_mix={"product": 1.0},  # type: ignore[arg-type]
        )


def test_piecewise_family_without_enabled_component_raises_when_mix_is_present() -> None:
    g = _make_generator(6)
    x = torch.randn(32, 4, generator=g)
    with pytest.raises(
        ValueError,
        match="enables 'piecewise' but disables all piecewise component",
    ):
        apply_random_function(
            x,
            g,
            out_dim=2,
            function_type="piecewise",
            function_family_mix={"piecewise": 1.0},  # type: ignore[arg-type]
        )


def test_deterministic_with_family_mix() -> None:
    x = torch.randn(32, 4)
    mix = {"nn": 0.5, "linear": 0.5}
    y1 = apply_random_function(
        x.clone(),
        _make_generator(6),
        out_dim=3,
        mechanism_logit_tilt=0.7,
        function_family_mix=mix,  # type: ignore[arg-type]
    )
    y2 = apply_random_function(
        x.clone(),
        _make_generator(6),
        out_dim=3,
        mechanism_logit_tilt=0.7,
        function_family_mix=mix,  # type: ignore[arg-type]
    )
    torch.testing.assert_close(y1, y2)


def test_default_sampled_family_never_emits_piecewise() -> None:
    g = _make_generator(16)
    for _ in range(128):
        sampled = sample_function_family(
            g,
            mechanism_logit_tilt=1.0,
            function_family_mix=None,
        )
        assert sampled != "piecewise"


def test_sampled_family_can_emit_piecewise_when_mix_enables_it() -> None:
    g = _make_generator(17)
    mix = {"piecewise": 1.0}
    for _ in range(16):
        sampled = sample_function_family(
            g,
            mechanism_logit_tilt=0.8,
            function_family_mix=mix,  # type: ignore[arg-type]
        )
        assert sampled == "piecewise"


def test_piecewise_explicit_family_with_mix_is_deterministic() -> None:
    x = torch.randn(64, 4, generator=_make_generator(18))
    mix = {"piecewise": 0.5, "linear": 0.5}
    y1 = apply_random_function(
        x.clone(),
        _make_generator(19),
        out_dim=3,
        function_type="piecewise",
        function_family_mix=mix,  # type: ignore[arg-type]
    )
    y2 = apply_random_function(
        x.clone(),
        _make_generator(19),
        out_dim=3,
        function_type="piecewise",
        function_family_mix=mix,  # type: ignore[arg-type]
    )
    torch.testing.assert_close(y1, y2)


def test_random_functions_module_retains_family_sampling_patch_point() -> None:
    g = _make_generator(21)
    sampled = _sample_function_family(
        g,
        mechanism_logit_tilt=0.0,
        function_family_mix={"linear": 1.0},  # type: ignore[arg-type]
    )
    assert sampled == "linear"



"""Stable identity helpers for persisted request, corpus, and dataset artifacts."""

from __future__ import annotations

import hashlib
import json
import math
from collections.abc import Mapping
from typing import Any

from dagzoo.core.fixed_layout.prepare import normalize_fixed_layout_target_cells
from dagzoo.math import sanitize_json


def stable_blake2s_hex(payload: Mapping[str, Any], *, digest_size: int = 16) -> str:
    """Return a stable BLAKE2s hex digest for sanitized JSON-compatible payloads."""

    encoded = json.dumps(
        sanitize_json(dict(payload)),
        sort_keys=True,
        separators=(",", ":"),
        allow_nan=False,
    ).encode("utf-8")
    return hashlib.blake2s(encoded, digest_size=int(digest_size)).hexdigest()


def _require_mapping(value: object, *, path: str) -> Mapping[str, Any]:
    if not isinstance(value, Mapping):
        raise ValueError(f"{path} must be a mapping.")
    return value


def _require_string(value: object, *, path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"{path} must be a non-empty string.")
    return value


def _require_int(value: object, *, path: str) -> int:
    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"{path} must be an integer.")
    return int(value)


def _require_optional_int(value: object, *, path: str) -> int | None:
    if value is None:
        return None
    return _require_int(value, path=path)


def _require_number(value: object, *, path: str) -> float:
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        raise ValueError(f"{path} must be a finite number.")
    number = float(value)
    if not math.isfinite(number):
        raise ValueError(f"{path} must be a finite number.")
    return number


def _normalized_float_mapping(value: object, *, path: str) -> dict[str, float]:
    mapping = _require_mapping(value, path=path)
    normalized: dict[str, float] = {}
    for key in sorted(mapping):
        normalized[str(key)] = _require_number(mapping[key], path=f"{path}.{key}")
    return normalized


def canonical_request_run_provenance(metadata: Mapping[str, Any]) -> dict[str, Any]:
    """Return the canonical run-level provenance payload used for dataset identities."""

    config = _require_mapping(metadata.get("config"), path="metadata.config")
    dataset = _require_mapping(config.get("dataset"), path="metadata.config.dataset")
    runtime = _require_mapping(config.get("runtime"), path="metadata.config.runtime")
    noise_distribution = _require_mapping(
        metadata.get("noise_distribution"),
        path="metadata.noise_distribution",
    )
    shift = _require_mapping(metadata.get("shift"), path="metadata.shift")

    missing_rate = _require_number(
        dataset.get("missing_rate"),
        path="metadata.config.dataset.missing_rate",
    )
    missing_mechanism = _require_string(
        dataset.get("missing_mechanism"),
        path="metadata.config.dataset.missing_mechanism",
    )
    if missing_rate <= 0.0:
        missing_mechanism = "none"
    missingness_payload: dict[str, Any] = {
        "missing_rate": missing_rate,
        "missing_mechanism": missing_mechanism,
    }
    if missing_mechanism == "mar":
        missingness_payload["missing_mar_observed_fraction"] = _require_number(
            dataset.get("missing_mar_observed_fraction"),
            path="metadata.config.dataset.missing_mar_observed_fraction",
        )
        missingness_payload["missing_mar_logit_scale"] = _require_number(
            dataset.get("missing_mar_logit_scale"),
            path="metadata.config.dataset.missing_mar_logit_scale",
        )
    elif missing_mechanism == "mnar":
        missingness_payload["missing_mnar_logit_scale"] = _require_number(
            dataset.get("missing_mnar_logit_scale"),
            path="metadata.config.dataset.missing_mnar_logit_scale",
        )

    noise_family = _require_string(
        noise_distribution.get("family_requested"),
        path="metadata.noise_distribution.family_requested",
    )
    noise_payload: dict[str, Any] = {
        "family_requested": noise_family,
        "base_scale": _require_number(
            noise_distribution.get("base_scale"),
            path="metadata.noise_distribution.base_scale",
        ),
    }
    if noise_family == "student_t":
        noise_payload["student_t_df"] = _require_number(
            noise_distribution.get("student_t_df"),
            path="metadata.noise_distribution.student_t_df",
        )
    elif noise_family == "mixture":
        noise_payload["mixture_weights"] = _normalized_float_mapping(
            noise_distribution.get("mixture_weights"),
            path="metadata.noise_distribution.mixture_weights",
        )

    return {
        "dataset": {
            "task": _require_string(
                dataset.get("task"),
                path="metadata.config.dataset.task",
            ),
            "n_train": _require_int(
                dataset.get("n_train"),
                path="metadata.config.dataset.n_train",
            ),
            "n_test": _require_int(
                dataset.get("n_test"),
                path="metadata.config.dataset.n_test",
            ),
            "missingness": missingness_payload,
        },
        "noise": noise_payload,
        "shift": {
            "variance_sigma_multiplier": _require_number(
                shift.get("variance_sigma_multiplier"),
                path="metadata.shift.variance_sigma_multiplier",
            )
        },
        "runtime": {
            "fixed_layout_target_cells": normalize_fixed_layout_target_cells(
                _require_optional_int(
                    runtime.get("fixed_layout_target_cells"),
                    path="metadata.config.runtime.fixed_layout_target_cells",
                )
            ),
            "resolved_device": _require_string(
                metadata.get("resolved_device"),
                path="metadata.resolved_device",
            ),
            "compute_backend": _require_string(
                metadata.get("compute_backend"),
                path="metadata.compute_backend",
            ),
            "torch_dtype": _require_string(
                runtime.get("torch_dtype"),
                path="metadata.config.runtime.torch_dtype",
            ),
        },
    }


def canonical_request_run_split_group(
    *,
    seed: int,
    run_num_datasets: int,
    layout_signature: str,
    layout_plan_signature: str,
    request_run_provenance: Mapping[str, Any],
) -> str:
    """Return a stable split-group key for one canonical request run."""

    return stable_blake2s_hex(
        {
            "seed": int(seed),
            "run_num_datasets": int(run_num_datasets),
            "layout_signature": str(layout_signature),
            "layout_plan_signature": str(layout_plan_signature),
            "request_run_provenance": dict(request_run_provenance),
        }
    )


def canonical_layout_plan_split_group(
    *,
    layout_signature: str,
    layout_plan_signature: str,
    layout_execution_contract: str,
) -> str:
    """Return a stable split-group key for one fixed-layout execution plan."""

    return stable_blake2s_hex(
        {
            "layout_signature": str(layout_signature),
            "layout_plan_signature": str(layout_plan_signature),
            "layout_execution_contract": str(layout_execution_contract),
        }
    )


def canonical_dataset_id(
    *,
    request_run_split_group: str,
    layout_plan_split_group: str,
    dataset_index: int,
    dataset_seed: int,
) -> str:
    """Return a stable dataset identifier for one canonical dataset bundle."""

    return stable_blake2s_hex(
        {
            "request_run_split_group": str(request_run_split_group),
            "layout_plan_split_group": str(layout_plan_split_group),
            "dataset_index": int(dataset_index),
            "dataset_seed": int(dataset_seed),
        }
    )


__all__ = [
    "canonical_dataset_id",
    "canonical_layout_plan_split_group",
    "canonical_request_run_provenance",
    "canonical_request_run_split_group",
    "stable_blake2s_hex",
]



"""Threshold-calibration helpers built on the diversity-audit engine."""

from __future__ import annotations

import math
from typing import Any, Sequence

from dagzoo.config import GeneratorConfig, clone_generator_config
from dagzoo.filter_thresholds import (
    FILTER_THRESHOLD_MAX,
    FILTER_THRESHOLD_MIN,
    validate_filter_threshold,
)

from .runner import run_effective_diversity_audit

DEFAULT_FILTER_CALIBRATION_DELTAS: tuple[float, ...] = (-0.15, -0.10, -0.05, 0.0, 0.05)
_CALIBRATION_THRESHOLD_MIN = FILTER_THRESHOLD_MIN
_CALIBRATION_THRESHOLD_MAX = FILTER_THRESHOLD_MAX


def validate_filter_calibration_threshold(
    value: object,
    *,
    field_name: str,
) -> float:
    """Validate one calibration threshold value."""

    return validate_filter_threshold(value, field_name=field_name)


def _normalize_threshold_candidate(value: float) -> float:
    """Normalize one requested threshold to a stable persisted float."""

    return round(float(value), 6)


def _clamp_threshold_candidate(value: float) -> float:
    """Clamp one requested threshold into the default calibration sweep window."""

    return min(_CALIBRATION_THRESHOLD_MAX, max(_CALIBRATION_THRESHOLD_MIN, float(value)))


def _threshold_label(value: float) -> str:
    """Return a stable variant label for one requested threshold."""

    return f"thr_{format_filter_calibration_threshold(value)}"


def format_filter_calibration_threshold(value: object) -> str:
    """Render a threshold value with stable precision for user-facing surfaces."""

    if isinstance(value, bool) or not isinstance(value, (int, float)) or not math.isfinite(value):
        return "-"
    normalized = _normalize_threshold_candidate(float(value))
    rendered = f"{normalized:.6f}".rstrip("0").rstrip(".")
    if "." not in rendered:
        rendered = f"{rendered}.0"
    return rendered


def resolve_filter_calibration_thresholds(
    *,
    baseline_threshold: float,
    thresholds: Sequence[float] | None,
) -> list[float]:
    """Resolve the requested threshold sweep including the baseline threshold."""

    baseline_value = _normalize_threshold_candidate(
        validate_filter_calibration_threshold(
            baseline_threshold,
            field_name="baseline_threshold",
        )
    )
    if thresholds is None:
        raw_candidates = [
            _clamp_threshold_candidate(float(baseline_value) + float(delta))
            for delta in DEFAULT_FILTER_CALIBRATION_DELTAS
        ]
    else:
        raw_candidates = [
            _normalize_threshold_candidate(
                validate_filter_calibration_threshold(
                    value,
                    field_name="thresholds",
                )
            )
            for value in thresholds
        ]

    normalized = {_normalize_threshold_candidate(float(value)) for value in raw_candidates}
    normalized.add(baseline_value)
    return sorted(normalized)


def _build_threshold_variant(config: GeneratorConfig, *, threshold: float) -> GeneratorConfig:
    """Clone one config and override only the requested filter threshold."""

    variant = clone_generator_config(config, revalidate=False)
    variant.filter.threshold = float(threshold)
    return variant


def _candidate_entry(
    entry: dict[str, Any],
    *,
    threshold_requested: float,
    diversity_status: str,
    comparison: dict[str, Any] | None,
) -> dict[str, Any]:
    """Build one flattened calibration candidate record."""

    filter_summary = entry.get("filter_summary")
    if not isinstance(filter_summary, dict):
        filter_summary = {}
    comparison_payload = comparison if isinstance(comparison, dict) else {}
    return {
        "label": entry.get("label"),
        "config_path": entry.get("config_path"),
        "suite": entry.get("suite"),
        "num_datasets": entry.get("num_datasets"),
        "warmup_datasets": entry.get("warmup_datasets"),
        "requested_device": entry.get("requested_device"),
        "resolved_device": entry.get("resolved_device"),
        "resolved_config": entry.get("resolved_config"),
        "threshold_requested": float(threshold_requested),
        "threshold_effective_mean": filter_summary.get("threshold_effective_mean"),
        "threshold_delta_mean": filter_summary.get("threshold_delta_mean"),
        "wins_ratio_mean": filter_summary.get("wins_ratio_mean"),
        "n_valid_oob_mean": filter_summary.get("n_valid_oob_mean"),
        "accepted_true_fraction": filter_summary.get("accepted_true_fraction"),
        "reason_counts": dict(filter_summary.get("reason_counts", {})),
        "datasets_per_minute": entry.get("datasets_per_minute"),
        "filter_datasets_per_minute": entry.get("filter_datasets_per_minute"),
        "filter_accepted_datasets_per_minute": entry.get("filter_accepted_datasets_per_minute"),
        "filter_accepted_datasets_measured": entry.get("filter_accepted_datasets_measured"),
        "filter_rejected_datasets_measured": entry.get("filter_rejected_datasets_measured"),
        "filter_acceptance_rate_dataset_level": entry.get("filter_acceptance_rate_dataset_level"),
        "filter_rejection_rate_dataset_level": entry.get("filter_rejection_rate_dataset_level"),
        "mechanism_family_summary": entry.get("mechanism_family_summary"),
        "diversity_status": str(diversity_status),
        "diversity_composite_shift_pct": comparison_payload.get("diversity_composite_shift_pct"),
        "diversity_metric_shift_pct": comparison_payload.get("diversity_metric_shift_pct"),
        "datasets_per_minute_delta_pct": comparison_payload.get("datasets_per_minute_delta_pct"),
        "filter_accepted_datasets_per_minute_delta_pct": comparison_payload.get(
            "filter_accepted_datasets_per_minute_delta_pct"
        ),
    }


def _ranking_value(candidate: dict[str, Any]) -> float:
    """Return ranking score for candidate selection."""

    value = candidate.get("filter_accepted_datasets_per_minute")
    if isinstance(value, bool) or not isinstance(value, (int, float)) or not math.isfinite(value):
        return float("-inf")
    return float(value)


def run_filter_calibration(
    *,
    config: GeneratorConfig,
    config_path: str,
    thresholds: Sequence[float] | None,
    suite: str,
    num_datasets: int | None,
    warmup: int | None,
    device: str | None,
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> dict[str, Any]:
    """Run threshold-only filter calibration against the rewritten audit engine."""

    if not bool(config.filter.enabled):
        raise ValueError("filter-calibration requires filter.enabled: true in the resolved config.")

    baseline_threshold = _normalize_threshold_candidate(
        validate_filter_calibration_threshold(
            config.filter.threshold,
            field_name="filter.threshold",
        )
    )
    threshold_candidates = resolve_filter_calibration_thresholds(
        baseline_threshold=baseline_threshold,
        thresholds=thresholds,
    )
    variant_thresholds = [value for value in threshold_candidates if value != baseline_threshold]
    variant_labels = [_threshold_label(value) for value in variant_thresholds]
    audit_report = run_effective_diversity_audit(
        baseline_config=config,
        baseline_config_path=config_path,
        variant_configs=[
            _build_threshold_variant(config, threshold=value) for value in variant_thresholds
        ],
        variant_config_paths=[config_path] * len(variant_thresholds),
        variant_labels=variant_labels,
        suite=suite,
        num_datasets=num_datasets,
        warmup=warmup,
        device=device,
        warn_threshold_pct=warn_threshold_pct,
        fail_threshold_pct=fail_threshold_pct,
    )

    baseline_entry = audit_report["baseline"]
    variant_entries = audit_report.get("variants", [])
    comparisons = audit_report.get("comparisons", [])
    if len(variant_entries) != len(variant_thresholds):
        raise ValueError("audit_report variants length did not match calibration thresholds.")
    if len(comparisons) != len(variant_thresholds):
        raise ValueError("audit_report comparisons length did not match calibration thresholds.")

    baseline_candidate = _candidate_entry(
        baseline_entry,
        threshold_requested=baseline_threshold,
        diversity_status="reference",
        comparison=None,
    )
    candidates = [baseline_candidate]
    for threshold, variant_entry, comparison in zip(
        variant_thresholds,
        variant_entries,
        comparisons,
        strict=True,
    ):
        candidates.append(
            _candidate_entry(
                variant_entry,
                threshold_requested=threshold,
                diversity_status=str(
                    comparison.get("diversity_status", "insufficient_metrics")
                    if isinstance(comparison, dict)
                    else "insufficient_metrics"
                ),
                comparison=comparison,
            )
        )
    candidates.sort(
        key=lambda item: (
            float(item.get("threshold_requested", 0.0)),
            str(item.get("label", "")),
        )
    )

    best_overall = max(candidates, key=_ranking_value)
    passing_candidates = [
        candidate for candidate in candidates if str(candidate.get("diversity_status")) == "pass"
    ]
    best_passing = max(passing_candidates, key=_ranking_value) if passing_candidates else None
    overall_status = (
        str(audit_report.get("summary", {}).get("overall_status", "insufficient_metrics"))
        if comparisons
        else "reference"
    )

    return {
        "schema_name": "dagzoo_filter_calibration_report",
        "schema_version": 1,
        "generated_at": audit_report.get("generated_at"),
        "baseline": baseline_candidate,
        "candidates": candidates,
        "comparisons": comparisons,
        "summary": {
            "overall_status": overall_status,
            "warn_threshold_pct": float(warn_threshold_pct),
            "fail_threshold_pct": float(fail_threshold_pct),
            "num_candidates": len(candidates),
            "baseline_threshold_requested": float(baseline_threshold),
            "best_overall_threshold_requested": float(best_overall["threshold_requested"]),
            "best_overall_diversity_status": str(best_overall["diversity_status"]),
            "best_passing_threshold_requested": (
                float(best_passing["threshold_requested"]) if best_passing is not None else None
            ),
            "probe_num_datasets": audit_report.get("summary", {}).get("probe_num_datasets"),
            "probe_warmup_datasets": audit_report.get("summary", {}).get("probe_warmup_datasets"),
            "threshold_candidates_requested": [
                float(candidate["threshold_requested"]) for candidate in candidates
            ],
        },
    }



"""Effective-config rendering and persistence helpers."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml

from dagzoo.config import GeneratorConfig


def effective_config_yaml(config: GeneratorConfig) -> str:
    """Render an effective config payload as YAML text."""

    return yaml.safe_dump(
        config.to_dict(),
        sort_keys=False,
        default_flow_style=False,
    )


def write_effective_config(config: GeneratorConfig, path: Path) -> Path:
    """Persist effective config YAML to disk."""

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(effective_config_yaml(config), encoding="utf-8")
    return path


def effective_resolution_trace_yaml(trace_payload: list[dict[str, Any]]) -> str:
    """Render a field-level config resolution trace as YAML text."""

    return yaml.safe_dump(
        trace_payload,
        sort_keys=False,
        default_flow_style=False,
    )


def write_effective_config_trace(trace_payload: list[dict[str, Any]], path: Path) -> Path:
    """Persist effective config resolution trace YAML to disk."""

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(effective_resolution_trace_yaml(trace_payload), encoding="utf-8")
    return path


def print_effective_config(config: GeneratorConfig, *, header: str) -> None:
    """Print effective config YAML to stdout with a short header."""

    print(header)
    print(effective_config_yaml(config).rstrip())


def print_resolution_trace(trace_payload: list[dict[str, Any]], *, header: str) -> None:
    """Print config resolution trace YAML to stdout with a short header."""

    print(header)
    print(effective_resolution_trace_yaml(trace_payload).rstrip())



#!/usr/bin/env python3
"""Check repo-root Markdown path references for stale or missing targets."""

from __future__ import annotations

import argparse
import re
import sys
from pathlib import Path
from typing import Iterable

REPO_ROOT = Path(__file__).resolve().parents[2]
DEFAULT_ROOTS = (
    "README.md",
    "docs",
    "scripts/README.md",
)
REPO_PATH_PREFIXES = (
    "src/",
    "docs/",
    "scripts/",
    "configs/",
    "reference/",
    "site/",
)
REPO_PATH_EXACT = frozenset(
    {
        "README.md",
        "CHANGELOG.md",
        "AGENTS.md",
        "pyproject.toml",
        ".pre-commit-config.yaml",
    }
)
LEGACY_PATH_DENYLIST = frozenset(
    {
        "src/dagzoo/cli.py",
        "src/dagzoo/config.py",
        "src/dagzoo/meta_targets.py",
        "src/dagzoo/core/parallel_generation.py",
        "src/dagzoo/core/worker_partition.py",
        "src/dagzoo/core/generation_engine.py",
    }
)
INLINE_CODE_RE = re.compile(r"`([^`\n]+)`")
MARKDOWN_LINK_RE = re.compile(r"!?\[[^\]\n]*\]\(([^)\n]+)\)")
URI_SCHEME_RE = re.compile(r"^[a-zA-Z][a-zA-Z0-9+.-]*:")


def _iter_markdown_files(root: Path) -> Iterable[Path]:
    if root.is_file():
        if root.suffix.lower() == ".md":
            yield root
        return
    for path in root.rglob("*.md"):
        if path.is_file():
            yield path


def _iter_inline_code_spans(path: Path) -> Iterable[tuple[int, str]]:
    in_fence = False
    for lineno, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        stripped = line.strip()
        if stripped.startswith("```"):
            in_fence = not in_fence
            continue
        if in_fence:
            continue
        for match in INLINE_CODE_RE.finditer(line):
            yield lineno, match.group(1).strip()


def _iter_markdown_link_targets(path: Path) -> Iterable[tuple[int, str]]:
    in_fence = False
    for lineno, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        stripped = line.strip()
        if stripped.startswith("```"):
            in_fence = not in_fence
            continue
        if in_fence:
            continue
        line_without_code = INLINE_CODE_RE.sub("", line)
        for match in MARKDOWN_LINK_RE.finditer(line_without_code):
            yield lineno, match.group(1).strip()


def _is_repo_path(candidate: str) -> bool:
    return candidate in REPO_PATH_EXACT or candidate.startswith(REPO_PATH_PREFIXES)


def _normalize_repo_path_candidate(candidate: str) -> str | None:
    stripped = candidate.strip().rstrip(".,:;")
    if any(char in stripped for char in "*?[]{}"):
        return None
    if (REPO_ROOT / stripped).exists():
        return stripped
    token = stripped.split()[0]
    if any(char in token for char in "*?[]{}"):
        return None
    return token


def _normalize_markdown_link_candidate(candidate: str, *, source_path: Path) -> str | None:
    stripped = candidate.strip()
    if not stripped:
        return None
    if stripped.startswith("<") and stripped.endswith(">"):
        stripped = stripped[1:-1].strip()
    if not stripped:
        return None
    if URI_SCHEME_RE.match(stripped):
        return None
    if stripped.startswith(("#", "/")):
        return None

    link_target = stripped.split(maxsplit=1)[0]
    path_only = re.split(r"[?#]", link_target, maxsplit=1)[0]
    if not path_only or any(char in path_only for char in "*?[]{}"):
        return None

    resolved = (source_path.parent / path_only).resolve(strict=False)
    if not resolved.is_relative_to(REPO_ROOT):
        return None
    return resolved.relative_to(REPO_ROOT).as_posix()


def _scan_file(path: Path) -> list[tuple[int, str]]:
    errors: list[tuple[int, str]] = []
    for lineno, candidate in _iter_inline_code_spans(path):
        normalized = _normalize_repo_path_candidate(candidate)
        if normalized is None or not _is_repo_path(normalized):
            continue
        if normalized in LEGACY_PATH_DENYLIST:
            errors.append((lineno, f"{normalized} (legacy path)"))
            continue
        if not (REPO_ROOT / normalized).exists():
            errors.append((lineno, f"{normalized} (missing path)"))
    for lineno, candidate in _iter_markdown_link_targets(path):
        normalized = _normalize_markdown_link_candidate(candidate, source_path=path)
        if normalized is None or not _is_repo_path(normalized):
            continue
        if normalized in LEGACY_PATH_DENYLIST:
            errors.append((lineno, f"{normalized} (legacy path)"))
            continue
        if not (REPO_ROOT / normalized).exists():
            errors.append((lineno, f"{normalized} (missing path)"))
    return errors


def parse_args(argv: Iterable[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "roots",
        nargs="*",
        default=DEFAULT_ROOTS,
        help="Repo-relative Markdown file or directory roots to scan.",
    )
    return parser.parse_args(list(argv))


def main(argv: Iterable[str] | None = None) -> int:
    args = parse_args(sys.argv[1:] if argv is None else argv)
    missing_roots: list[str] = []
    all_errors: list[tuple[Path, int, str]] = []
    for root_rel in args.roots:
        root = REPO_ROOT / root_rel
        if not root.exists():
            missing_roots.append(root_rel)
            continue
        for path in _iter_markdown_files(root):
            for lineno, target in _scan_file(path):
                all_errors.append((path, lineno, target))

    if missing_roots or all_errors:
        if missing_roots:
            print("Missing Markdown scan roots:")
            for root_rel in missing_roots:
                print(f"- {root_rel}")
        if all_errors:
            print("Broken repo path references found:")
            for path, lineno, target in all_errors:
                print(f"- {path.relative_to(REPO_ROOT)}:{lineno} -> {target}")
        return 1

    print("Repo path check passed.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Numeric converter implementations."""

from __future__ import annotations

from types import SimpleNamespace

import torch

from dagzoo.core.execution_semantics import sample_converter_plan
from dagzoo.core.fixed_layout.batched import FixedLayoutBatchRng, apply_numeric_converter_plan_batch
from dagzoo.core.fixed_layout.plan_types import NumericConverterPlan
from dagzoo.rng import keyed_rng_from_generator


def apply_numeric_converter(
    x: torch.Tensor,
    generator: torch.Generator,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Numeric converter in torch via the shared typed-plan semantics."""
    y = x.to(torch.float32)
    if y.dim() == 1:
        y = y.unsqueeze(1)

    root = keyed_rng_from_generator(generator, "apply_numeric_converter")
    plan = sample_converter_plan(
        SimpleNamespace(key="value", kind="num", dim=int(y.shape[1]), cardinality=None),
        keyed_rng=root.keyed("plan"),
        mechanism_logit_tilt=0.0,
        function_family_mix=None,
        device=str(generator.device),
    )
    if not isinstance(plan, NumericConverterPlan):
        raise RuntimeError("Expected numeric converter plan for numeric converter.")

    rng = FixedLayoutBatchRng.from_keyed_rng(
        root.keyed("execution"),
        batch_size=1,
        device=str(y.device),
    )
    x_prime, values = apply_numeric_converter_plan_batch(
        y.unsqueeze(0),
        rng,
        plan,
    )
    return x_prime.squeeze(0), values.squeeze(0)



"""Shared generation finalization helpers for canonical fixed-layout execution."""

from __future__ import annotations

import copy
from dataclasses import asdict, dataclass
from typing import Any

import torch

from dagzoo.config import MISSINGNESS_MECHANISM_NONE, GeneratorConfig
from dagzoo.core.layout_types import LayoutPlan
from dagzoo.core.metadata import _build_lineage_metadata, _build_shift_metadata
from dagzoo.core.noise_runtime import (
    NoiseRuntimeSelection,
    _build_noise_distribution_metadata,
)
from dagzoo.core.shift import ShiftRuntimeParams
from dagzoo.core.validation import (
    InfeasibleStratifiedSplitError,
    InvalidClassSplitError,
    _classification_split_valid,
    _stratified_split_indices,
)
from dagzoo.postprocess.postprocess import (
    inject_missingness,
    postprocess_dataset,
    postprocess_fixed_schema_batch,
)
from dagzoo.rng import KeyedRng
from dagzoo.types import DatasetBundle


@dataclass(slots=True)
class _FixedSchemaFinalizationContext:
    """Cached metadata used by fixed-schema chunk finalization."""

    config_payload: dict[str, Any]
    shift_metadata: dict[str, Any]
    feature_types: list[str]
    feature_index_map: list[int]
    base_metadata: dict[str, Any]
    missingness_enabled: bool


def _config_payload_for_metadata(
    config: GeneratorConfig,
    *,
    n_train: int,
    n_test: int,
) -> dict[str, Any]:
    """Serialize config metadata while omitting unset fixed-layout target-cell overrides."""

    config_payload = asdict(config)
    dataset_payload = config_payload.get("dataset")
    if isinstance(dataset_payload, dict):
        dataset_payload["n_train"] = int(n_train)
        dataset_payload["n_test"] = int(n_test)

    runtime_payload = config_payload.get("runtime")
    if (
        isinstance(runtime_payload, dict)
        and runtime_payload.get("fixed_layout_target_cells") is None
    ):
        runtime_payload.pop("fixed_layout_target_cells", None)
    return config_payload


def _classification_class_structure(
    *,
    y_train: torch.Tensor,
    y_test: torch.Tensor,
    n_classes_sampled: int,
) -> dict[str, Any]:
    """Build classification label-structure metadata for one emitted bundle."""

    y_train_i64 = y_train.to(torch.int64)
    y_test_i64 = y_test.to(torch.int64)
    y_all = torch.cat([y_train_i64, y_test_i64], dim=0)
    unique_all = torch.unique(y_all, sorted=True)
    n_classes_realized = int(unique_all.numel())
    labels_contiguous = bool(
        torch.equal(
            unique_all,
            torch.arange(n_classes_realized, dtype=unique_all.dtype, device=unique_all.device),
        )
    )
    train_classes = torch.unique(y_train_i64, sorted=True)
    test_classes = torch.unique(y_test_i64, sorted=True)

    return {
        "n_classes_sampled": int(n_classes_sampled),
        "n_classes_realized": int(n_classes_realized),
        "labels_contiguous": bool(labels_contiguous),
        "train_test_class_match": bool(torch.equal(train_classes, test_classes)),
        "min_label": int(unique_all[0].item()) if n_classes_realized > 0 else None,
        "max_label": int(unique_all[-1].item()) if n_classes_realized > 0 else None,
    }


def _resolve_split_indices(
    y: torch.Tensor,
    *,
    task: str,
    n_train: int,
    keyed_rng: KeyedRng,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Resolve one dataset's train/test split indices with CUDA preference and CPU fallback."""

    device_candidates = ("cuda", "cpu") if str(y.device.type) == "cuda" else ("cpu",)
    last_error: Exception | None = None
    for split_device in device_candidates:
        try:
            return _resolve_split_indices_for_device(
                y,
                task=task,
                n_train=n_train,
                keyed_rng=keyed_rng,
                device=split_device,
            )
        except InfeasibleStratifiedSplitError:
            raise
        except Exception as exc:
            last_error = exc
            continue
    if last_error is not None:
        raise last_error
    raise RuntimeError("Failed to resolve train/test split indices.")


def _resolve_split_indices_for_device(
    y: torch.Tensor,
    *,
    task: str,
    n_train: int,
    keyed_rng: KeyedRng,
    device: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Resolve one dataset's train/test split indices on one requested split device."""

    generator = keyed_rng.torch_rng(device=device)
    split_y = y if str(y.device.type) == device else y.to(device=device)
    if task == "classification":
        return _stratified_split_indices(
            split_y,
            n_train,
            generator,
            device,
        )

    total_rows = int(split_y.shape[0])
    order = torch.randperm(
        total_rows,
        generator=generator,
        device=device,
    )
    return order[:n_train], order[n_train:]


def _split_raw_tensors(
    x: torch.Tensor,
    y: torch.Tensor,
    *,
    train_idx: torch.Tensor,
    test_idx: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Split one raw feature/target pair using indices applied on the tensor device."""

    train_index = train_idx if train_idx.device == x.device else train_idx.to(device=x.device)
    test_index = test_idx if test_idx.device == x.device else test_idx.to(device=x.device)
    return x[train_index], y[train_index], x[test_index], y[test_index]


def _normalized_filter_metadata(aux_meta: dict[str, Any]) -> dict[str, Any]:
    """Return the emitted filter metadata shape for one dataset."""

    filter_metadata = aux_meta.get("filter", {})
    if isinstance(filter_metadata, dict):
        return dict(filter_metadata)
    return {"mode": "deferred", "status": "not_run"}


def _is_missingness_enabled(config: GeneratorConfig) -> bool:
    """Return whether missingness injection can mutate emitted feature tensors."""

    return (
        float(config.dataset.missing_rate) > 0.0
        and str(config.dataset.missing_mechanism).strip().lower() != MISSINGNESS_MECHANISM_NONE
    )


def _base_bundle_metadata_for_layout(
    layout: LayoutPlan,
    *,
    feature_types: list[str],
    feature_index_map: list[int],
) -> dict[str, Any]:
    """Build invariant metadata shared across bundles with one emitted feature schema."""

    return {
        "backend": "torch",
        "compute_backend": "torch_appendix_full",
        "n_features": int(len(feature_types)),
        "n_categorical_features": int(sum(1 for t in feature_types if t == "cat")),
        "graph_nodes": int(layout.graph_nodes),
        "graph_edges": int(layout.graph_edges),
        "graph_depth_nodes": int(layout.graph_depth_nodes),
        "graph_edge_density": float(layout.graph_edge_density),
        "lineage": _build_lineage_metadata(layout, feature_index_map=feature_index_map),
    }


def _build_fixed_schema_finalization_context(
    config: GeneratorConfig,
    layout: LayoutPlan,
    *,
    n_train: int,
    n_test: int,
    shift_params: ShiftRuntimeParams,
    feature_types: list[str] | None = None,
    feature_index_map: list[int] | None = None,
) -> _FixedSchemaFinalizationContext:
    """Build cached metadata for fixed-schema bundle finalization."""

    config_payload = _config_payload_for_metadata(
        config,
        n_train=n_train,
        n_test=n_test,
    )
    resolved_feature_types = (
        [str(feature_type) for feature_type in list(layout.feature_types)]
        if feature_types is None
        else [str(feature_type) for feature_type in feature_types]
    )
    resolved_feature_index_map = (
        [int(i) for i in range(int(layout.n_features))]
        if feature_index_map is None
        else [int(i) for i in feature_index_map]
    )
    return _FixedSchemaFinalizationContext(
        config_payload=config_payload,
        shift_metadata=_build_shift_metadata(
            shift_params=shift_params,
            function_family_mix=config.mechanism.function_family_mix,
        ),
        feature_types=resolved_feature_types,
        feature_index_map=resolved_feature_index_map,
        base_metadata=_base_bundle_metadata_for_layout(
            layout,
            feature_types=resolved_feature_types,
            feature_index_map=resolved_feature_index_map,
        ),
        missingness_enabled=_is_missingness_enabled(config),
    )


def _build_bundle_metadata(
    context: _FixedSchemaFinalizationContext,
    *,
    seed: int,
    attempt: int,
    attempts_used: int,
    device: str,
    requested_device: str,
    resolved_device: str,
    device_fallback_reason: str | None,
    aux_meta: dict[str, Any],
    noise_runtime_selection: NoiseRuntimeSelection,
    missingness_summary: dict[str, Any] | None,
    class_structure: dict[str, Any] | None,
) -> dict[str, Any]:
    """Build emitted dataset metadata for scalar and batched finalization paths."""

    n_classes = None if class_structure is None else int(class_structure["n_classes_realized"])
    metadata = dict(context.base_metadata)
    metadata["lineage"] = copy.deepcopy(context.base_metadata["lineage"])
    metadata.update(
        {
            "device": str(device),
            "requested_device": str(requested_device),
            "resolved_device": str(resolved_device),
            "device_fallback_reason": device_fallback_reason,
            "n_classes": n_classes,
            "seed": int(seed),
            "attempt_used": int(attempt),
            "filter": _normalized_filter_metadata(aux_meta),
            "shift": copy.deepcopy(context.shift_metadata),
            "noise_distribution": _build_noise_distribution_metadata(noise_runtime_selection),
            "generation_attempts": {
                "total_attempts": int(attempts_used),
                "retry_count": int(max(0, attempts_used - 1)),
                "filter_attempts": 0,
                "filter_rejections": 0,
                "filter_rejection_rate": None,
            },
            "config": copy.deepcopy(context.config_payload),
        }
    )
    if missingness_summary is not None:
        metadata["missingness"] = missingness_summary
    if class_structure is not None:
        metadata["class_structure"] = class_structure
    return metadata


def _finalize_processed_bundle(
    config: GeneratorConfig,
    layout: LayoutPlan,
    *,
    context: _FixedSchemaFinalizationContext,
    dataset_seed: int,
    attempt: int,
    attempts_used: int,
    attempt_root: KeyedRng,
    device: str,
    requested_device: str,
    resolved_device: str,
    device_fallback_reason: str | None,
    x_train: torch.Tensor,
    y_train: torch.Tensor,
    x_test: torch.Tensor,
    y_test: torch.Tensor,
    aux_meta: dict[str, Any],
    noise_runtime_selection: NoiseRuntimeSelection,
    dtype: torch.dtype,
) -> DatasetBundle:
    """Finalize one already-postprocessed dataset into the emitted bundle contract."""

    if context.missingness_enabled:
        x_train, x_test, missingness_summary = inject_missingness(
            x_train,
            x_test,
            dataset_cfg=config.dataset,
            keyed_rng=attempt_root.keyed("missingness"),
            device=device,
        )
    else:
        missingness_summary = None

    if config.dataset.task == "classification" and not _classification_split_valid(y_train, y_test):
        raise InvalidClassSplitError("invalid_class_split")

    x_train = x_train.to(device=device, dtype=dtype)
    x_test = x_test.to(device=device, dtype=dtype)
    y_dtype = torch.int64 if config.dataset.task == "classification" else dtype
    y_train = y_train.to(device=device, dtype=y_dtype)
    y_test = y_test.to(device=device, dtype=y_dtype)

    class_structure: dict[str, Any] | None = None
    if config.dataset.task == "classification":
        class_structure = _classification_class_structure(
            y_train=y_train,
            y_test=y_test,
            n_classes_sampled=int(layout.n_classes),
        )

    metadata = _build_bundle_metadata(
        context,
        seed=dataset_seed,
        attempt=attempt,
        attempts_used=attempts_used,
        device=device,
        requested_device=requested_device,
        resolved_device=resolved_device,
        device_fallback_reason=device_fallback_reason,
        aux_meta=aux_meta,
        noise_runtime_selection=noise_runtime_selection,
        missingness_summary=missingness_summary,
        class_structure=class_structure,
    )
    return DatasetBundle(
        X_train=x_train,
        y_train=y_train,
        X_test=x_test,
        y_test=y_test,
        feature_types=list(context.feature_types),
        metadata=metadata,
        runtime_metrics={},
    )


def _finalize_generated_chunk_preserve_schema(
    config: GeneratorConfig,
    layout: LayoutPlan,
    *,
    context: _FixedSchemaFinalizationContext,
    dataset_roots: list[KeyedRng],
    attempt: int,
    attempts_used: int,
    device: str,
    n_train: int,
    n_test: int,
    requested_device: str,
    resolved_device: str,
    device_fallback_reason: str | None,
    x: torch.Tensor,
    y: torch.Tensor,
    aux_meta_batch: list[dict[str, Any]],
    noise_runtime_selection: NoiseRuntimeSelection,
    dtype: torch.dtype,
    resolved_split_indices: list[tuple[torch.Tensor, torch.Tensor] | None] | None = None,
) -> list[DatasetBundle | None]:
    """Finalize one fixed-schema raw chunk while preserving scalar retry semantics."""

    if int(x.shape[0]) != len(dataset_roots) or int(y.shape[0]) != len(dataset_roots):
        raise ValueError("Chunk tensors must align with provided dataset roots.")
    if resolved_split_indices is not None and len(resolved_split_indices) != len(dataset_roots):
        raise ValueError("resolved_split_indices must align with provided dataset roots.")

    results: list[DatasetBundle | None] = [None] * len(dataset_roots)
    valid_positions: list[int] = []
    train_idx_list: list[torch.Tensor] = []
    test_idx_list: list[torch.Tensor] = []
    postprocess_roots: list[KeyedRng] = []
    for batch_index, dataset_root in enumerate(dataset_roots):
        if resolved_split_indices is not None:
            split_indices = resolved_split_indices[batch_index]
            if split_indices is None:
                continue
            train_idx, test_idx = split_indices
        else:
            attempt_root = dataset_root.keyed("attempt", attempt)
            try:
                train_idx, test_idx = _resolve_split_indices(
                    y[batch_index],
                    task=config.dataset.task,
                    n_train=n_train,
                    keyed_rng=attempt_root.keyed("split"),
                )
            except InfeasibleStratifiedSplitError:
                continue

        valid_positions.append(int(batch_index))
        train_idx_list.append(train_idx)
        test_idx_list.append(test_idx)
        postprocess_roots.append(dataset_root.keyed("attempt", attempt, "postprocess"))

    if not valid_positions:
        return results

    valid_index = torch.as_tensor(valid_positions, dtype=torch.long, device=x.device)
    x_valid = x.index_select(0, valid_index)
    y_valid = y.index_select(0, valid_index)
    train_idx = torch.stack([idx.to(device=x.device) for idx in train_idx_list])
    test_idx = torch.stack([idx.to(device=x.device) for idx in test_idx_list])

    x_train_t = torch.gather(
        x_valid,
        1,
        train_idx.unsqueeze(-1).expand(-1, -1, int(x.shape[2])),
    )
    x_test_t = torch.gather(
        x_valid,
        1,
        test_idx.unsqueeze(-1).expand(-1, -1, int(x.shape[2])),
    )
    y_train_t = torch.gather(y_valid, 1, train_idx)
    y_test_t = torch.gather(y_valid, 1, test_idx)

    x_train, y_train, x_test, y_test = postprocess_fixed_schema_batch(
        x_train_t,
        y_train_t,
        x_test_t,
        y_test_t,
        list(context.feature_types),
        config.dataset.task,
        postprocess_roots=postprocess_roots,
    )
    for local_index, batch_index in enumerate(valid_positions):
        try:
            dataset_root = dataset_roots[batch_index]
            dataset_seed = dataset_root.child_seed()
            results[batch_index] = _finalize_processed_bundle(
                config,
                layout,
                context=context,
                dataset_seed=dataset_seed,
                attempt=attempt,
                attempts_used=attempts_used,
                attempt_root=dataset_root.keyed("attempt", attempt),
                device=device,
                requested_device=requested_device,
                resolved_device=resolved_device,
                device_fallback_reason=device_fallback_reason,
                x_train=x_train[local_index],
                y_train=y_train[local_index],
                x_test=x_test[local_index],
                y_test=y_test[local_index],
                aux_meta=aux_meta_batch[batch_index],
                noise_runtime_selection=noise_runtime_selection,
                dtype=dtype,
            )
        except InvalidClassSplitError:
            continue

    return results


def _finalize_generated_tensors(
    config: GeneratorConfig,
    layout: LayoutPlan,
    *,
    dataset_seed: int,
    attempt: int,
    attempts_used: int,
    dataset_root: KeyedRng,
    device: str,
    n_train: int,
    n_test: int,
    requested_device: str,
    resolved_device: str,
    device_fallback_reason: str | None,
    x: torch.Tensor,
    y: torch.Tensor,
    aux_meta: dict[str, Any],
    shift_params: ShiftRuntimeParams,
    noise_runtime_selection: NoiseRuntimeSelection,
    dtype: torch.dtype,
    preserve_feature_schema: bool = False,
    finalization_context: _FixedSchemaFinalizationContext | None = None,
) -> DatasetBundle:
    """Finalize one raw `x`/`y` pair into the standard dataset bundle contract."""

    attempt_root = dataset_root.keyed("attempt", attempt)
    try:
        train_idx, test_idx = _resolve_split_indices(
            y,
            task=config.dataset.task,
            n_train=n_train,
            keyed_rng=attempt_root.keyed("split"),
        )
    except InfeasibleStratifiedSplitError as exc:
        raise InvalidClassSplitError("invalid_class_split") from exc

    x_train_t, y_train_t, x_test_t, y_test_t = _split_raw_tensors(
        x,
        y,
        train_idx=train_idx,
        test_idx=test_idx,
    )

    x_train, y_train, x_test, y_test, feature_types, feature_index_map = postprocess_dataset(
        x_train_t,
        y_train_t,
        x_test_t,
        y_test_t,
        list(layout.feature_types),
        config.dataset.task,
        attempt_root.keyed("postprocess"),
        device,
        return_feature_index_map=True,
        preserve_feature_schema=preserve_feature_schema,
    )
    if (
        finalization_context is not None
        and list(feature_types) == list(finalization_context.feature_types)
        and [int(i) for i in feature_index_map] == list(finalization_context.feature_index_map)
    ):
        context = finalization_context
    else:
        context = _build_fixed_schema_finalization_context(
            config,
            layout,
            n_train=n_train,
            n_test=n_test,
            shift_params=shift_params,
            feature_types=feature_types,
            feature_index_map=feature_index_map,
        )
    return _finalize_processed_bundle(
        config,
        layout,
        context=context,
        dataset_seed=dataset_seed,
        attempt=attempt,
        attempts_used=attempts_used,
        attempt_root=attempt_root,
        device=device,
        requested_device=requested_device,
        resolved_device=resolved_device,
        device_fallback_reason=device_fallback_reason,
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        aux_meta=aux_meta,
        noise_runtime_selection=noise_runtime_selection,
        dtype=dtype,
    )
